from __future__ import annotations

import json
import os
from pathlib import Path
import subprocess
import sys
from types import SimpleNamespace

import pytest

from gotta import builtin as plugin_api
import gotta.content.context as content_context
import gotta.content.env as content_env
import gotta.content.model as content_model
import gotta.content.scope as content_scope
import gotta.content.store as content_store
import gotta.cli.argv as cli_argv
import gotta.cli.entry as cli
import gotta.cli.notice as cli_notice
import gotta.dispatch.budget as dispatch_budget
import gotta.dispatch.main as dispatch
import gotta.dispatch.materialize as dispatch_materialize
import gotta.resolve.invoke as invocation
from gotta.actor import ACTOR_ID_ENV
from gotta.capture import Capture
from gotta.plugins import read as read_plugin
from gotta.plugins.session import main as session_plugin
from gotta.resolve.search import SearchRouteError, resolve_search_route


def test_should_materialize_respects_help_and_suppression(monkeypatch) -> None:
    assert not dispatch.should_materialize("read", ["--help"])
    assert not dispatch.should_materialize(
        "read", ["artifact:demo.md@abc123", "--head", "20"]
    )
    assert dispatch.should_materialize(
        "read", ["https://github.com/acme/widgets", "--head", "3"]
    )
    assert dispatch.should_materialize(
        "read", ["https://example.com/manual.txt", "--tail", "5"]
    )
    assert not dispatch.should_materialize("session", [])
    assert not dispatch.should_materialize("session", ["show"])
    assert not dispatch.should_materialize("session", ["analyze"])
    assert not dispatch.should_materialize("want", [])
    assert not dispatch.should_materialize("goal", ["--help"])
    assert not dispatch.should_materialize("gdocs", ["status"])
    assert dispatch.should_materialize("confluence", ["search", "abc"])
    assert dispatch.should_materialize("confluence", ["get", "10101"])
    assert dispatch.should_materialize("gdocs", ["search", "abc"])
    assert dispatch.should_materialize("gdocs", ["get", "doc-123"])
    assert dispatch.should_materialize("gdrive", ["search", "abc"])
    assert dispatch.should_materialize("gdrive", ["get", "file-123"])
    assert dispatch.should_materialize("grafana", ["search", "--type", "dash-db"])
    assert dispatch.should_materialize("grafana", ["search", "abc"])
    assert dispatch.should_materialize("grafana", ["get", "dash-123"])
    assert not dispatch.should_materialize(
        "grafana", ["query", "--datasource", "prom-main", "sum(up)"]
    )
    assert dispatch.should_materialize("granola", ["search", "abc"])
    assert dispatch.should_materialize("granola", ["get", "note-123"])
    assert dispatch.should_materialize("gsheets", ["search", "abc"])
    assert dispatch.should_materialize("gsheets", ["get", "sheet-123"])
    assert dispatch.should_materialize("github", ["search", "abc"])
    assert dispatch.should_materialize("github", ["https://github.com/acme/widgets"])
    assert dispatch.should_materialize("jira", ["search", "abc"])
    assert dispatch.should_materialize("jira", ["get", "PROJ-1"])
    assert dispatch.should_materialize("slack", ["search", "abc"])
    assert dispatch.should_materialize("slack", ["get", "C12345678:1773085070.240949"])
    monkeypatch.setenv(dispatch_materialize.SUPPRESS_MATERIALIZATION_ENV, "1")
    assert not dispatch.should_materialize(
        "github", ["https://github.com/acme/widgets"]
    )


def test_split_common_options_strips_shared_actor_target() -> None:
    options, cleaned = dispatch.split_common_options(
        [
            "search",
            "platform",
            "--session",
            "demo",
            "--actor",
            "claude",
            "--save-as",
            "x.md",
        ],
        strip_actor=True,
    )

    assert options.session_dir == "demo"
    assert options.actor == "claude"
    assert options.save_as == "x.md"
    assert cleaned == ["search", "platform"]


def test_search_plugin_spec_exposes_unary_should_materialize_contract() -> None:
    spec = plugin_api.get_plugin("search")

    assert spec is not None
    assert spec.should_materialize is not None
    assert spec.should_materialize(["jira:Architecture"]) is True
    assert spec.should_materialize(["readme.md"]) is False


def test_emit_budgeted_output_truncates_interactive_text_with_footer(
    monkeypatch, capsys
) -> None:
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    payload = ("\n".join(f"line {index}" for index in range(400)) + "\n").encode(
        "utf-8"
    )

    emitted = dispatch.emit_budgeted_output(
        payload,
        output_format="text",
        budget_output=True,
        follow_command="gotta read artifact:demo@abc123",
    )
    captured = capsys.readouterr()

    assert emitted.output_truncated is True
    assert emitted.truncate_reason == "lines"
    assert "output truncated by lines budget" in captured.out
    assert "gotta read artifact:demo@abc123" in captured.out
    assert len(captured.out.encode("utf-8")) <= dispatch_budget.OUTPUT_EMIT_BYTE_LIMIT


def test_emit_budgeted_output_omits_overlong_follow_command_in_text_footer(
    monkeypatch, capsys
) -> None:
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    payload = ("\n".join(f"line {index}" for index in range(400)) + "\n").encode(
        "utf-8"
    )

    dispatch.emit_budgeted_output(
        payload,
        output_format="text",
        budget_output=True,
        follow_command="gotta read " + ("x" * 500),
    )
    captured = capsys.readouterr()

    assert "output truncated by lines budget" in captured.out
    assert "follow: " not in captured.out
    assert "rerun the same command with --full-output" in captured.out


def test_emit_budgeted_output_emits_json_preview_envelope_for_interactive_json(
    capsys,
) -> None:
    payload = {"items": [{"id": index, "text": "x" * 64} for index in range(500)]}

    emitted = dispatch.emit_budgeted_output(
        json.dumps(payload, indent=2, sort_keys=True).encode("utf-8"),
        output_format="json",
        budget_output=True,
        follow_command="gotta read artifact:demo@abc123",
    )
    rendered = json.loads(capsys.readouterr().out)

    assert emitted.output_truncated is True
    assert rendered["outputTruncated"] is True
    assert rendered["truncateReason"] == "bytes"
    assert rendered["followCommand"] == "gotta read artifact:demo@abc123"


def test_emit_budgeted_output_keeps_json_preview_valid_with_long_follow_command(
    capsys,
) -> None:
    payload = {"items": [{"id": index, "text": "x" * 64} for index in range(500)]}

    emitted = dispatch.emit_budgeted_output(
        json.dumps(payload, indent=2, sort_keys=True).encode("utf-8"),
        output_format="json",
        budget_output=True,
        follow_command="x" * 20000,
    )
    raw = capsys.readouterr().out.encode("utf-8")
    rendered = json.loads(raw)

    assert emitted.output_truncated is True
    assert len(raw) <= dispatch_budget.OUTPUT_BUDGET_BYTE_LIMIT
    assert rendered["outputTruncated"] is True
    assert rendered["requestedFormat"] == "json"
    assert rendered["truncateReason"] == "bytes"
    assert "followCommand" not in rendered
    assert (
        len(rendered.get("preview", "")) <= dispatch_budget.JSON_PREVIEW_CHAR_LIMIT + 3
    )


def test_select_text_cutoff_can_use_flex_to_finish_paragraph(monkeypatch) -> None:
    monkeypatch.setattr(dispatch_budget, "OUTPUT_BUDGET_FLEX_BYTE_LIMIT", 16)

    payload = ("A" * 71 + "\n\n" + "B" * 80).encode("utf-8")
    cutoff = dispatch_budget._select_text_cutoff(payload, soft_limit=70, hard_limit=86)

    assert cutoff == 73


def test_run_plugin_actor_launch_streams_live_without_buffered_capture(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    def fake_runner(_argv: list[str]) -> int:
        print("launch stdout")
        print("launch stderr", file=sys.stderr, flush=True)
        return 0

    def fake_resolve_invocation(
        _plugin: str, _argv: list[str], _options: content_model.CommonOptions
    ):
        return SimpleNamespace(should_materialize=False, artifact_intent="none")

    def forbidden_capture(*_args, **_kwargs):
        raise AssertionError("actor launch should bypass buffered capture")

    monkeypatch.setattr(dispatch, "load_plugin_runner", lambda _plugin: fake_runner)
    monkeypatch.setattr(dispatch, "resolve_invocation", fake_resolve_invocation)
    monkeypatch.setattr(dispatch, "session_access_mode", lambda _plugin, _argv: "none")
    monkeypatch.setattr(dispatch, "capture_stdout", forbidden_capture)
    monkeypatch.setattr(dispatch, "capture_stderr", forbidden_capture)

    assert (
        dispatch.run_plugin(
            "actor", ["launch", "helper", "--session", str(tmp_path / "session")]
        )
        == 0
    )
    captured = capsys.readouterr()

    assert "launch stdout" in captured.out
    assert "launch stderr" in captured.err


def test_session_access_mode_tracks_artifact_bearing_surfaces() -> None:
    assert dispatch.session_access_mode("jira", ["search", "platform"]) == "ambient"
    assert dispatch.session_access_mode("jira", ["status"]) == "none"
    assert dispatch.session_access_mode("confluence", ["get", "10101"]) == "ambient"
    assert (
        dispatch.session_access_mode("confluence", ["replace", "10101", "a", "b"])
        == "none"
    )
    assert dispatch.session_access_mode("read", ["README.md"]) == "ambient"
    assert dispatch.session_access_mode("search", ["jira:platform"]) == "ambient"


def test_search_resolve_invocation_routes_provider_search_with_implicit_search() -> (
    None
):
    resolved = invocation.resolve_invocation(
        "search", ["jira:Architecture"], content_model.CommonOptions()
    )

    assert resolved.entry_plugin == "search"
    assert resolved.resolved_plugin == "jira"
    assert resolved.resolved_argv == ["search", "Architecture"]
    assert resolved.artifact_intent == "discovery"
    assert resolved.should_materialize is True


def test_search_resolve_invocation_accepts_explicit_search_alias() -> None:
    resolved = invocation.resolve_invocation(
        "search",
        ["slack:search ABC reboot"],
        content_model.CommonOptions(),
    )

    assert resolved.resolved_plugin == "slack"
    assert resolved.resolved_argv == ["search", "ABC reboot"]


def test_search_resolve_route_rejects_extra_unquoted_query_terms() -> None:
    with pytest.raises(
        SearchRouteError,
        match=r"top-level `gotta search` takes exactly one provider-qualified plain-text query string; quote the full `<provider>:<query>` argument or use `gotta github search \.\.\.` for structured flags",
    ):
        resolve_search_route(["github:SomeFunction", "ownership"])


def test_search_resolve_route_rejects_unknown_flag_shaped_tokens() -> None:
    with pytest.raises(
        SearchRouteError,
        match=r"top-level `gotta search` takes exactly one provider-qualified plain-text query string; quote the full `<provider>:<query>` argument or use `gotta jira search \.\.\.` for structured flags",
    ):
        resolve_search_route(["jira:retry", "--bogus", "bar"])


def test_search_resolve_route_preserves_flag_shaped_text_inside_quoted_query() -> None:
    route = resolve_search_route(["jira:retry --bogus bar"])

    assert route.provider == "jira"
    assert route.provider_argv == ["search", "retry --bogus bar"]


def test_search_resolve_invocation_disables_materialization_on_invalid_target() -> None:
    resolved = invocation.resolve_invocation(
        "search", ["jira:jql project = OPS"], content_model.CommonOptions()
    )

    assert resolved.should_materialize is False
    assert resolved.artifact_intent == "none"


@pytest.mark.parametrize(
    (
        "argv",
        "expected_plugin",
        "expected_args",
        "expected_intent",
        "expected_kind",
        "expected_materialize",
    ),
    [
        (
            ["github:search platform"],
            "github",
            ["search", "platform"],
            "discovery",
            "discovery",
            True,
        ),
        (
            ["granola:list"],
            "granola",
            ["list"],
            "discovery",
            "discovery",
            True,
        ),
        (
            ["slack:workspace:demo"],
            "slack",
            ["status", "--workspace", "demo", "--output", "summary"],
            "none",
            "",
            False,
        ),
        (
            ["grafana:status"],
            "grafana",
            ["status"],
            "none",
            "",
            False,
        ),
    ],
)
def test_read_resolve_invocation_preserves_routed_provider_artifact_intent(
    argv: list[str],
    expected_plugin: str,
    expected_args: list[str],
    expected_intent: str,
    expected_kind: str,
    expected_materialize: bool,
) -> None:
    resolved = dispatch.resolve_invocation("read", argv, content_model.CommonOptions())

    assert resolved.resolved_plugin == expected_plugin
    assert resolved.resolved_argv == expected_args
    assert resolved.artifact_intent == expected_intent
    assert resolved.artifact_kind == expected_kind
    assert resolved.should_materialize is expected_materialize


def test_derive_preferred_name_for_delegated_read_targets() -> None:
    options = content_model.CommonOptions()
    cases = {
        "https://github.com/acme/widgets#readme": "widgets.json",
        "https://github.com/acme/widgets/commits/main": "widgets-commits-main.json",
        "https://github.com/acme/widgets/pull/19": "widgets-pr-19.json",
        "granola:11111111-1111-1111-1111-111111111111": "11111111-1111-1111-1111-111111111111.json",
        "github:search --type pr --repo acme/widgets ABC": "github-search-prs-acme-widgets-abc.json",
        "https://example.atlassian.net/wiki/spaces/ENG/pages/10101/Platform+Architecture+Overview": "10101.html",
        "https://example.atlassian.net/wiki/pages/viewpage.action?pageId=20202": "20202.html",
        "https://example.atlassian.net/wiki/x/1J0AAA": "40404.html",
        "https://example.atlassian.net/browse/PROJ-3960": "PROJ-3960.json",
        "https://drive.google.com/file/d/drive-file-123/view?usp=sharing": "drive-file-123.bin",
        "https://docs.google.com/document/d/doc-123/edit?usp=drivesdk": "doc-123.html",
        "https://docs.google.com/spreadsheets/d/sheet-123/edit#gid=0": "sheet-123.json",
        "https://example.slack.com/archives/C12345678/p1773085070240949": "p1773085070240949.json",
        "slack:example-workspace": "slack-workspace-example-workspace.summary",
    }

    for target, expected in cases.items():
        assert dispatch.derive_preferred_name("read", [target], options) == expected


def test_derive_preferred_name_for_provider_search_artifacts() -> None:
    options = content_model.CommonOptions()

    assert (
        dispatch.derive_preferred_name("confluence", ["search", "ABC"], options)
        == "confluence-search-abc.json"
    )
    assert (
        dispatch.derive_preferred_name("gdocs", ["search", "ABC"], options)
        == "gdocs-search-abc.json"
    )
    assert (
        dispatch.derive_preferred_name("gdrive", ["search", "ABC"], options)
        == "gdrive-search-abc.json"
    )
    assert (
        dispatch.derive_preferred_name("granola", ["search", "ABC"], options)
        == "granola-search-abc.json"
    )
    assert (
        dispatch.derive_preferred_name(
            "granola",
            ["list", "--sort", "created", "--order", "asc", "--offset", "10"],
            options,
        )
        == "granola-list-created-asc-offset-10.json"
    )
    assert (
        dispatch.derive_preferred_name("gsheets", ["search", "ABC"], options)
        == "gsheets-search-abc.json"
    )
    assert (
        dispatch.derive_preferred_name(
            "jira",
            ["search", "--limit", "25", "--output", "markdown", "ABCC OR ABCS"],
            options,
        )
        == "jira-search-abcc-or-abcs.json"
    )
    assert (
        dispatch.derive_preferred_name(
            "slack",
            [
                "search",
                "--workspace",
                "example-workspace",
                "--source",
                "archive",
                "ABC",
            ],
            options,
        )
        == "slack-search-example-workspace-abc.json"
    )
    assert (
        dispatch.derive_preferred_name(
            "slack",
            ["status", "--workspace", "example-workspace", "--output", "summary"],
            options,
        )
        == "slack-workspace-example-workspace.summary"
    )
    assert (
        dispatch.derive_preferred_name(
            "grafana",
            ["search", "--type", "dash-db", "prod"],
            options,
        )
        == "grafana-search-dash-db-prod.json"
    )
    assert (
        dispatch.derive_preferred_name(
            "grafana",
            ["search", "Production Overview"],
            options,
        )
        == "grafana-search-production-overview.json"
    )
    assert (
        dispatch.derive_preferred_name(
            "grafana",
            ["query", "--datasource", "prom-main", "sum(up)"],
            options,
        )
        == "grafana-query-sum-up.summary"
    )


def test_derive_preferred_name_for_provider_get_artifacts_with_flags() -> None:
    options = content_model.CommonOptions()

    assert (
        dispatch.derive_preferred_name(
            "confluence",
            ["get", "--output", "markdown", "10101"],
            options,
        )
        == "10101.html"
    )
    assert (
        dispatch.derive_preferred_name(
            "jira",
            ["get", "--output", "markdown", "PROJ-3960"],
            options,
        )
        == "PROJ-3960.json"
    )
    assert (
        dispatch.derive_preferred_name(
            "gdocs",
            [
                "get",
                "--output",
                "markdown",
                "https://docs.google.com/document/d/doc-123/edit",
            ],
            options,
        )
        == "doc-123.html"
    )
    assert (
        dispatch.derive_preferred_name(
            "gdrive",
            ["get", "--output", "markdown", "file-123"],
            options,
        )
        == "file-123.bin"
    )
    assert (
        dispatch.derive_preferred_name(
            "slack",
            [
                "get",
                "--workspace",
                "example-workspace",
                "--output",
                "markdown",
                "https://example.slack.com/archives/C12345678/p1773085070240949",
            ],
            options,
        )
        == "p1773085070240949.json"
    )


def test_root_help_exposes_session_aware_read_storage_contract(
    capsys, monkeypatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(content_scope, "DEFAULT_SESSION_ROOT", tmp_path / "session")

    assert cli.main(["--help"]) == 0

    output = capsys.readouterr().err
    assert (
        "acquire one target through the right retrieval surface with session-aware storage"
        in output
    )


def test_require_operational_session_accepts_initialized_session(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path,
        content_dir=tmp_path / "content",
    )
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)
    dispatch.require_operational_session(dirs)


def test_require_operational_session_requires_initialized_session(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path,
        content_dir=tmp_path / "content",
    )
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    with pytest.raises(content_model.ContentError, match="gotta"):
        dispatch.require_operational_session(dirs)


def test_canonical_locator_normalizes_common_provider_shapes() -> None:
    assert (
        dispatch.canonical_locator(
            "read",
            ["https://github.com/acme/widgets/tree/main/docs/ARCHITECTURE.md"],
        )
        == "https://github.com/acme/widgets/tree/main/docs/ARCHITECTURE.md"
    )
    assert (
        dispatch.canonical_locator(
            "read",
            ["https://github.com/acme/widgets#readme"],
        )
        == "https://github.com/acme/widgets"
    )
    assert (
        dispatch.canonical_locator(
            "granola",
            ["get", "11111111-1111-1111-1111-111111111111"],
        )
        == "granola:11111111-1111-1111-1111-111111111111"
    )
    assert (
        dispatch.canonical_locator(
            "granola",
            ["get", "Weekly Review"],
        )
        == "granola:get 'Weekly Review'"
    )
    assert (
        dispatch.canonical_locator(
            "granola",
            ["search", "--limit", "5", "latency"],
        )
        == "granola:search --limit 5 latency"
    )
    assert (
        dispatch.canonical_locator(
            "granola",
            ["transcript", "Weekly Review", "--query", "latency"],
        )
        == "granola:transcript 'Weekly Review' --query latency"
    )
    assert (
        dispatch.canonical_locator(
            "jira",
            [
                "get",
                "--output",
                "markdown",
                "https://example.atlassian.net/browse/PROJ-3960",
            ],
        )
        == "jira:PROJ-3960"
    )
    assert (
        dispatch.canonical_locator(
            "jira",
            ["get", "https://example.atlassian.net/browse/PROJ-3960"],
        )
        == "jira:PROJ-3960"
    )
    assert (
        dispatch.canonical_locator(
            "confluence",
            [
                "get",
                "--output",
                "markdown",
                "https://example.atlassian.net/wiki/spaces/ENG/pages/10101/Page",
            ],
        )
        == "confluence:10101"
    )
    assert (
        dispatch.canonical_locator(
            "grafana",
            ["get", "demo-dashboard-uid"],
        )
        == "grafana:get demo-dashboard-uid"
    )
    assert (
        dispatch.canonical_locator(
            "confluence",
            [
                "get",
                "https://example.atlassian.net/wiki/spaces/ENG/pages/10101/Page",
            ],
        )
        == "confluence:10101"
    )
    assert (
        dispatch.canonical_locator(
            "confluence",
            [
                "get",
                "https://example.atlassian.net/wiki/pages/viewpage.action?pageId=20202",
            ],
        )
        == "confluence:20202"
    )
    assert (
        dispatch.canonical_locator(
            "read",
            [
                "https://example.atlassian.net/wiki/spaces/ENG/pages/10101/Page"
                "?focusedCommentId=30303",
            ],
        )
        == "confluence:30303"
    )
    assert (
        dispatch.canonical_locator(
            "confluence",
            ["get", "https://example.atlassian.net/wiki/x/1J0AAA"],
        )
        == "confluence:40404"
    )
    assert (
        dispatch.canonical_locator(
            "confluence",
            ["get", "https://example.atlassian.net/wiki/x/GoD9AgE"],
        )
        == "confluence:4345135130"
    )
    assert (
        dispatch.canonical_locator(
            "slack",
            [
                "get",
                "--workspace",
                "example-workspace",
                "--output",
                "markdown",
                "https://example.slack.com/archives/C12345678/p1773085070240949",
            ],
        )
        == "slack:thread:C12345678:1773085070240949"
    )
    assert (
        dispatch.canonical_locator(
            "slack",
            ["get", "https://example.slack.com/archives/C12345678/p1773085070240949"],
        )
        == "slack:thread:C12345678:1773085070240949"
    )
    assert (
        dispatch.canonical_locator(
            "slack",
            ["get", "https://example.slack.com/docs/T12345678/F12345678"],
        )
        == "slack:doc:T12345678:F12345678"
    )
    assert (
        dispatch.canonical_locator(
            "read",
            ["https://example.slack.com/docs/T12345678/F12345678"],
        )
        == "slack:doc:T12345678:F12345678"
    )
    assert (
        dispatch.canonical_locator(
            "gdrive",
            ["get", "https://drive.google.com/file/d/drive-file-123/view?usp=sharing"],
        )
        == "gdrive:drive-file-123"
    )
    assert (
        dispatch.canonical_locator(
            "gdocs",
            [
                "get",
                "--output",
                "markdown",
                "https://docs.google.com/document/d/doc-123/edit",
            ],
        )
        == "gdocs:doc-123"
    )
    assert (
        dispatch.canonical_locator(
            "gdocs",
            ["get", "https://docs.google.com/document/d/doc-123/edit"],
        )
        == "gdocs:doc-123"
    )
    assert (
        dispatch.canonical_locator(
            "gsheets",
            ["get", "https://docs.google.com/spreadsheets/d/sheet-123/edit#gid=0"],
        )
        == "gsheets:sheet-123"
    )
    assert (
        dispatch.canonical_locator(
            "read",
            ["https://docs.google.com/spreadsheets/d/sheet-123/edit#gid=0"],
        )
        == "gsheets:sheet-123"
    )
    assert (
        dispatch.canonical_locator(
            "read",
            [
                "https://github.com/acme/widgets/actions/runs/123456789/job/987654321#summary"
            ],
        )
        == "https://github.com/acme/widgets/actions/runs/123456789/job/987654321"
    )
    assert (
        dispatch.canonical_locator("jira", ["search", "Architecture"])
        == "jira:search Architecture"
    )
    assert (
        dispatch.canonical_locator("jira", ["issue-types", "--project", "OPS"])
        == "jira:issue-types --project OPS"
    )
    assert (
        dispatch.canonical_locator(
            "jira",
            ["projects", "--limit", "25", "--offset", "25"],
        )
        == "jira:projects --limit 25 --offset 25"
    )
    assert (
        dispatch.canonical_locator("jira", ["sprints", "--project", "OPS"])
        == "jira:sprints --project OPS"
    )
    assert (
        dispatch.canonical_locator(
            "grafana",
            ["datasources", "--limit", "25", "--offset", "25"],
        )
        == "grafana:datasources --limit 25 --offset 25"
    )
    assert (
        dispatch.canonical_locator(
            "jira",
            ["fields", "--project", "OPS", "--type", "Service Request"],
        )
        == "jira:fields --project OPS --type 'Service Request'"
    )
    assert (
        dispatch.canonical_locator(
            "jira",
            ["add-to-sprint", "OPS-42", "--current", "--project", "OPS"],
        )
        == "jira:add-to-sprint OPS-42 --current --project OPS"
    )
    assert (
        dispatch.canonical_locator("jira", ["transitions", "OPS-42"])
        == "jira:transitions OPS-42"
    )
    assert (
        dispatch.canonical_locator("confluence", ["search", "Architecture"])
        == "confluence:search Architecture"
    )


def test_canonical_locator_normalizes_reordered_read_search_locators() -> None:
    first = dispatch.canonical_locator(
        "read",
        ["slack:search ABC reboot --workspace example-workspace --limit 10"],
    )
    second = dispatch.canonical_locator(
        "read",
        ["slack:search --workspace example-workspace --limit 10 ABC reboot"],
    )

    assert second == first
    assert first.startswith("slack:search --workspace example-workspace --limit 10 ")
    assert "ABC reboot" in first


def test_materialize_invocation_attributes_delegated_read_to_provider(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path, content_dir=tmp_path / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "read",
        ["jira:PROJ-1"],
        content_model.CommonOptions(),
        b"# PROJ-1\n\n- Created: 2026-03-10T12:00:00Z\n",
        dirs=dirs,
    )

    assert result is not None
    snapshot = next(item for item in content_store.scan_content_store(dirs.content_dir))
    assert snapshot.metadata["plugin"] == "jira"
    assert snapshot.metadata["entrypoint"] == "read"
    assert snapshot.metadata["provider"] == "jira"
    assert snapshot.metadata["canonical_locator"] == "jira:PROJ-1"


def test_external_plugin_can_shadow_builtin(monkeypatch) -> None:
    class Dist:
        def __init__(self, name: str) -> None:
            self.name = name

    class EntryPoint:
        def __init__(self, name: str, dist_name: str, payload) -> None:
            self.name = name
            self.dist = Dist(dist_name)
            self.value = f"{dist_name}:{name}"
            self._payload = payload

        def load(self):
            return self._payload

    builtin = plugin_api.PluginSpec(
        name="github",
        description="builtin",
        runner=lambda argv: 0,
    )
    shadow = plugin_api.PluginSpec(
        name="github",
        description="shadow",
        runner=lambda argv: 0,
    )

    plugin_api.clear_plugin_cache()
    try:
        monkeypatch.setattr(
            plugin_api,
            "entry_points",
            lambda group: [
                EntryPoint("github", "gotta", builtin),
                EntryPoint("github", "gotta-plugin-github", shadow),
            ],
        )

        discovered = plugin_api.discovered_plugins()

        assert discovered["github"].description == "shadow"
    finally:
        plugin_api.clear_plugin_cache()


def test_plugin_discovery_is_group_scoped(monkeypatch) -> None:
    class Dist:
        def __init__(self, name: str) -> None:
            self.name = name

    class EntryPoint:
        def __init__(self, name: str, dist_name: str, payload) -> None:
            self.name = name
            self.dist = Dist(dist_name)
            self.value = f"{dist_name}:{name}"
            self._payload = payload

        def load(self):
            return self._payload

    monkeypatch.setattr(
        plugin_api,
        "entry_points",
        lambda group: (
            [
                EntryPoint(
                    "ask",
                    "gotta",
                    plugin_api.PluginSpec(
                        name="ask", description="", runner=lambda argv: 0
                    ),
                )
            ]
            if group == plugin_api.DEFAULT_PLUGIN_GROUP
            else [
                EntryPoint(
                    "docs",
                    "gotta-plugin-ask-docs",
                    plugin_api.PluginSpec(
                        name="docs", description="", runner=lambda argv: 0
                    ),
                )
            ]
        ),
    )

    plugin_api.clear_plugin_cache()
    try:
        plugins = plugin_api.available_plugins()
        assert "ask" in plugins
        assert "logs" in plugins
        assert "todo" in plugins
        assert plugin_api.available_plugins(group=plugin_api.ASK_PLUGIN_GROUP) == [
            "docs"
        ]
    finally:
        plugin_api.clear_plugin_cache()


def test_builtin_core_plugins_are_available_without_reinstalled_metadata() -> None:
    plugin_api.clear_plugin_cache()
    try:
        plugins = plugin_api.available_plugins()
        assert "goal" in plugins
        assert "logs" in plugins
        assert "todo" in plugins
        assert "want" in plugins
        assert plugin_api.get_plugin("goal") is not None
        assert plugin_api.get_plugin("logs") is not None
        assert plugin_api.get_plugin("todo") is not None
        assert plugin_api.get_plugin("want") is not None
    finally:
        plugin_api.clear_plugin_cache()


def test_source_seeded_core_plugins_ignore_stale_core_metadata(monkeypatch) -> None:
    class Dist:
        def __init__(self, name: str) -> None:
            self.name = name

    class EntryPoint:
        def __init__(self, name: str, dist_name: str, payload) -> None:
            self.name = name
            self.dist = Dist(dist_name)
            self.value = f"{dist_name}:{name}"
            self._payload = payload

        def load(self):
            return self._payload

    stale = plugin_api.PluginSpec(
        name="todo",
        description="stale installed metadata",
        runner=lambda argv: 0,
    )

    monkeypatch.setattr(
        plugin_api,
        "entry_points",
        lambda group: (
            [EntryPoint("todo", "gotta", stale)]
            if group == plugin_api.DEFAULT_PLUGIN_GROUP
            else []
        ),
    )

    plugin_api.clear_plugin_cache()
    try:
        assert (
            plugin_api.get_plugin("logs").description
            == "inspect and mutate the canonical session procedural trace"
        )
        assert (
            plugin_api.get_plugin("todo").description
            == "inspect and mutate the canonical session checklist"
        )
    finally:
        plugin_api.clear_plugin_cache()


def test_broken_external_ask_entry_points_do_not_break_help_all(
    monkeypatch, capsys
) -> None:
    class Dist:
        def __init__(self, name: str) -> None:
            self.name = name

    class EntryPoint:
        def __init__(self, name: str, dist_name: str) -> None:
            self.name = name
            self.dist = Dist(dist_name)
            self.value = f"{dist_name}:{name}"

        def load(self):
            raise ModuleNotFoundError("No module named 'gotta.ask.docs'")

    def fake_entry_points(group: str):
        if group == plugin_api.ASK_PLUGIN_GROUP:
            return [EntryPoint("docs", "gotta-plugin-ask-docs")]
        return []

    monkeypatch.setattr(plugin_api, "entry_points", fake_entry_points)

    plugin_api.clear_plugin_cache()
    try:
        assert plugin_api.available_plugins(group=plugin_api.ASK_PLUGIN_GROUP) == []
        assert cli.main(["--help-all"]) == 0
        captured = capsys.readouterr()
        output = captured.out
        error = captured.err
        assert "## gotta ask" in output
        assert "none installed" in output
        assert "warning: ignoring broken plugin entry point gotta.ask:docs" in error
        assert "ModuleNotFoundError" in error
    finally:
        plugin_api.clear_plugin_cache()


def test_cli_main_exits_zero_on_broken_pipe(monkeypatch) -> None:
    calls: list[str] = []

    monkeypatch.setattr(cli_argv, "available_plugins", lambda: ["read"])
    monkeypatch.setattr(cli_notice, "_silence_stdout", lambda: calls.append("silenced"))

    def raise_broken_pipe(plugin: str, argv: list[str]) -> int:
        raise BrokenPipeError()

    monkeypatch.setattr(cli_argv, "run_plugin", raise_broken_pipe)

    assert cli.main(["read", "README.md"]) == 0
    assert calls == ["silenced"]


def test_cli_help_all_includes_recursive_sections(capsys) -> None:
    assert cli.main(["--help-all"]) == 0
    captured = capsys.readouterr()
    output = captured.out + captured.err
    assert "# gotta" in output
    assert "Session synthesis surfaces live under `gotta session`" in output
    assert "`manifest`, `timeline`, `graph`, `leads`, `analyze`, `scan`" in output
    assert "This top-level long help shows only plugin root surfaces." in output
    assert (
        "Use `gotta <plugin> --help-all` for recursive help within one plugin."
        in output
    )
    assert "## gotta ask" in output
    assert "## gotta logs" in output
    assert "## gotta notes" in output
    assert "## gotta oops" in output
    assert "## gotta actor" in output
    assert "## gotta todo" in output
    assert "usage: gotta ask <surface> [args...]" in output
    assert "usage: gotta logs" in output
    assert "usage: gotta notes" in output
    assert "usage: gotta oops" in output
    assert "usage: gotta actor" in output
    assert "usage: gotta todo" in output
    assert "## gotta session" in output
    assert "scan" in output
    assert "Use --help-all for recursive command help." not in output
    assert "Use --help-all for the same long-form usage output." not in output
    assert "End of top-level long help for `gotta`." in output
    assert "Plugin subtrees were intentionally omitted at this level." in output


@pytest.mark.parametrize("plugin", ["slack", "jira", "confluence", "gdocs"])
def test_cli_plugin_help_all_works_from_top_level_dispatch(plugin: str, capsys) -> None:
    assert cli.main([plugin, "--help-all"]) == 0

    captured = capsys.readouterr()
    output = captured.out + captured.err
    assert f"## gotta {plugin}" in output
    assert "required: command" not in output


def test_cli_pipe_close_exits_cleanly_for_local_read_views(tmp_path: Path) -> None:
    root = tmp_path / "session-root"
    dirs = content_model.ResolvedDirs(
        session_dir=root,
        content_dir=root / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)

    large_file = tmp_path / "large.txt"
    large_file.write_text(
        "".join(
            f"line {index:06d} abcdefghijklmnopqrstuvwxyz\n" for index in range(200000)
        ),
        encoding="utf-8",
    )

    env = os.environ.copy()
    env[content_env.SESSION_ENV] = str(dirs.session_dir)

    proc = subprocess.Popen(
        [sys.executable, "-m", "gotta", "read", str(large_file)],
        cwd=tmp_path,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    assert proc.stdout is not None
    assert proc.stderr is not None

    first_line = proc.stdout.readline().rstrip("\n")
    proc.stdout.close()
    stderr = proc.stderr.read()
    returncode = proc.wait()

    assert first_line == "line 000000 abcdefghijklmnopqrstuvwxyz"
    assert returncode == 0
    assert "BrokenPipeError" not in stderr

    snapshots = content_store.scan_content_store(dirs.content_dir)
    assert snapshots == []


def test_run_plugin_local_read_does_not_emit_stored_content_receipt(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    dirs = content_model.ResolvedDirs(
        session_dir=root,
        content_dir=root / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)

    sample = tmp_path / "sample.txt"
    sample.write_text("hello\n", encoding="utf-8")

    monkeypatch.setenv(content_env.SESSION_ENV, str(dirs.session_dir))

    assert dispatch.run_plugin("read", [str(sample)]) == 0
    captured = capsys.readouterr()

    assert captured.out == "hello\n"
    assert captured.err == ""


def test_emit_budgeted_output_skips_default_budget_with_full_output_escape(
    capsys,
) -> None:
    payload = ("\n".join(f"line {index}" for index in range(400)) + "\n").encode(
        "utf-8"
    )

    emitted = dispatch.emit_budgeted_output(
        payload,
        output_format="text",
        budget_output=False,
        follow_command="gotta read artifact:demo@abc123",
    )
    captured = capsys.readouterr()

    assert emitted.output_budget_applied is False
    assert emitted.output_truncated is False
    assert captured.out == payload.decode("utf-8")


def test_run_plugin_read_section_miss_returns_clean_error_for_local_view(
    tmp_path: Path, capsys
) -> None:
    sample = tmp_path / "sample.md"
    sample.write_text("# Intro\n\nbody\n", encoding="utf-8")

    assert dispatch.run_plugin("read", [str(sample), "--section", "Missing"]) == 1
    captured = capsys.readouterr()

    assert captured.out == ""
    assert "no section heading matched: Missing" in captured.err
    assert "Traceback" not in captured.err


def test_run_plugin_materializing_read_section_miss_returns_clean_error(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    assert session_plugin.main(["init", "--session", str(local_root)]) == 0
    capsys.readouterr()

    monkeypatch.setattr(
        read_plugin,
        "fetch_url",
        lambda _target: (b"# Intro\n\nbody\n", "text/markdown", False),
    )

    assert (
        dispatch.run_plugin(
            "read",
            [
                "https://example.com/doc.md",
                "--section",
                "Missing",
                "--session",
                str(local_root),
            ],
        )
        == 1
    )
    captured = capsys.readouterr()

    assert captured.out == ""
    assert "no section heading matched: Missing" in captured.err
    assert "Traceback" not in captured.err
    assert content_store.scan_content_store(local_root / "content") == []


def test_run_plugin_session_scan_invalid_regex_fails_even_when_manifest_is_empty(
    tmp_path: Path, capsys
) -> None:
    local_root = tmp_path / "local"
    assert session_plugin.main(["init", "--session", str(local_root)]) == 0
    capsys.readouterr()

    assert (
        dispatch.run_plugin(
            "session",
            ["scan", "[", "--match", "regex", "--session", str(local_root)],
        )
        == 1
    )
    captured = capsys.readouterr()

    assert captured.out == ""
    assert "invalid scan pattern:" in captured.err
    assert "Traceback" not in captured.err


def test_run_plugin_read_invalid_confluence_shortlink_returns_clean_error(
    capsys,
) -> None:
    assert (
        dispatch.run_plugin("read", ["https://example.atlassian.net/wiki/x/!!!!!"]) == 1
    )
    captured = capsys.readouterr()

    assert captured.out == ""
    assert "could not parse Confluence page ID from input" in captured.err
    assert "Traceback" not in captured.err


def test_materialize_invocation_carries_slack_thread_source_timestamps(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "read",
        ["https://example.slack.com/archives/C12345678/p1773085070240949"],
        content_model.CommonOptions(),
        b"thread body\n",
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["source_created_at"] == "2026-03-09T19:37:50.240949Z"
    assert metadata["source_updated_at"] == "2026-03-09T19:37:50.240949Z"


def test_materialize_invocation_carries_slack_channel_source_window(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "slack",
        ["get", "https://example.slack.com/archives/C12345678", "--output", "meta"],
        content_model.CommonOptions(),
        json.dumps(
            {
                "firstTs": "1770935417.208289",
                "lastTs": "1773347711.122509",
            }
        ).encode("utf-8"),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["source_created_at"] == "2026-02-12T22:30:17.208289Z"
    assert metadata["source_updated_at"] == "2026-03-12T20:35:11.122509Z"


def test_materialize_invocation_extracts_slack_markdown_source_times(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "slack",
        [
            "get",
            "https://example.slack.com/archives/C12345678/p1773085070240949",
            "--output",
            "markdown",
        ],
        content_model.CommonOptions(),
        (
            b"### Example thread\n\n"
            b"- _Channel_: `#ops`\n"
            b"- _Source_: https://example.slack.com/archives/C12345678/p1773085070240949\n"
            b"- Created: 2026-03-09T19:37:50.240949Z\n"
            b"- Updated: 2026-03-09T20:01:11.000000Z\n\n"
            b"- _2026-03-09 14:37:50 CST_ **Alice**: example\n"
        ),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["source_created_at"] == "2026-03-09T19:37:50.240949Z"
    assert metadata["source_updated_at"] == "2026-03-09T20:01:11.000000Z"


def test_materialize_invocation_extracts_markdown_source_times(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "jira",
        ["get", "PROJ-1", "--output", "markdown"],
        content_model.CommonOptions(),
        (
            b"# PROJ-1: Example\n\n"
            b"- Created: 2026-03-10T12:00:00Z\n"
            b"- Updated: 2026-03-11T09:30:00Z\n"
        ),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["source_created_at"] == "2026-03-10T12:00:00Z"
    assert metadata["source_updated_at"] == "2026-03-11T09:30:00Z"


def test_materialize_invocation_extracts_json_source_times(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "github",
        ["https://github.com/example/repo/pull/1", "--output", "json"],
        content_model.CommonOptions(),
        json.dumps(
            {
                "createdAt": "2026-02-01T10:00:00Z",
                "updatedAt": "2026-02-03T11:30:00Z",
            }
        ).encode("utf-8"),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["source_created_at"] == "2026-02-01T10:00:00Z"
    assert metadata["source_updated_at"] == "2026-02-03T11:30:00Z"


def test_materialize_invocation_persists_visibility_metadata(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "github",
        ["https://github.com/example/repo", "--output", "json"],
        content_model.CommonOptions(),
        json.dumps(
            {
                "name": "repo",
                "url": "https://github.com/example/repo",
                "visibility": "private",
                "createdAt": "2026-02-01T10:00:00Z",
            }
        ).encode("utf-8"),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    manifest = json.loads(
        (dirs.content_dir / "manifest.jsonl").read_text(encoding="utf-8")
    )

    assert metadata["visibility_level"] == "restricted"
    assert metadata["visibility_boundary"] == "same_company"
    assert metadata["visibility_confidence"] == "high"
    assert metadata["visibility_basis"] == [
        "provider=github",
        "repo.visibility=private",
    ]
    assert manifest["visibility_level"] == "restricted"
    assert manifest["visibility_boundary"] == "same_company"
    assert manifest["visibility_confidence"] == "high"
    assert manifest["visibility_basis"] == [
        "provider=github",
        "repo.visibility=private",
    ]


def test_materialize_invocation_extracts_visibility_from_markdown(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "slack",
        [
            "get",
            "https://example.slack.com/archives/C12345678/p1773085070240949",
            "--output",
            "markdown",
        ],
        content_model.CommonOptions(),
        (
            b"### Slack Thread: Example\n\n"
            b"- _Channel_: `#ops`\n"
            b"- _Source_: https://example.slack.com/archives/C12345678/p1773085070240949\n"
            b"- Visibility: internal (same_company, high)\n"
            b"- Created: 2026-03-09T19:37:50.240949Z\n"
        ),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["visibility_level"] == "internal"
    assert metadata["visibility_boundary"] == "same_company"
    assert metadata["visibility_confidence"] == "high"


def test_materialize_invocation_derives_nested_search_source_times_from_json(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "gdocs",
        ["search", "ABC", "--output", "json"],
        content_model.CommonOptions(),
        json.dumps(
            {
                "query": "ABC",
                "results": [
                    {
                        "title": "Older doc",
                        "createdTime": "2026-01-01T10:00:00Z",
                        "modifiedTime": "2026-01-05T12:00:00Z",
                    },
                    {
                        "title": "Newer doc",
                        "createdTime": "2026-02-01T09:00:00Z",
                        "modifiedTime": "2026-03-07T08:30:00Z",
                    },
                ],
            }
        ).encode("utf-8"),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["source_created_at"] == "2026-01-01T10:00:00Z"
    assert metadata["source_updated_at"] == "2026-03-07T08:30:00Z"


def test_materialize_invocation_extracts_github_commit_history_markdown_source_range(
    tmp_path: Path,
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    result = dispatch._materialize_invocation(
        "github",
        ["https://github.com/example/repo/commits/main", "--output", "summary"],
        content_model.CommonOptions(),
        (
            b"# example/repo commit history for `main`\n\n"
            b"- **Created:** 2026-03-09T17:32:25Z\n"
            b"- **Updated:** 2026-03-11T02:14:24Z\n"
            b"- **Commits shown:** 14\n\n"
            b"- [abc1234](https://github.com/example/repo/commit/abc1234) Example\n"
        ),
        dirs=dirs,
    )

    assert result is not None
    metadata = json.loads(
        (dirs.content_dir / result.digest / "meta.json").read_text(encoding="utf-8")
    )
    assert metadata["source_created_at"] == "2026-03-09T17:32:25Z"
    assert metadata["source_updated_at"] == "2026-03-11T02:14:24Z"


def test_materialize_invocation_captures_actor_actor_metadata(
    tmp_path: Path, monkeypatch
) -> None:
    dirs = content_model.ResolvedDirs(
        session_dir=tmp_path / "session-root",
        content_dir=tmp_path / "session-root" / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)

    monkeypatch.setenv(ACTOR_ID_ENV, "claude")
    monkeypatch.setenv("GOTTA_ACTOR_DIR", str(dirs.session_dir / "actors" / "claude"))

    result = dispatch._materialize_invocation(
        "read",
        ["README.md"],
        content_model.CommonOptions(),
        b"hello\n",
        dirs=dirs,
    )

    assert result is not None
    manifest = json.loads(
        (dirs.content_dir / "manifest.jsonl").read_text(encoding="utf-8")
    )
    assert manifest["actor"] == "claude"
    assert manifest["actor_dir"].endswith("/actors/claude")


def test_materialize_invocation_rejects_unbound_actor_shell(
    tmp_path: Path, monkeypatch
) -> None:
    shared_root = tmp_path / "session-root"
    dirs = content_model.ResolvedDirs(
        session_dir=shared_root / "actors" / "claude",
        content_dir=shared_root / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)
    monkeypatch.delenv(ACTOR_ID_ENV, raising=False)
    monkeypatch.delenv(content_env.SESSION_ACTOR_ENV, raising=False)
    monkeypatch.delenv("GOTTA_ACTOR_SPEAKER", raising=False)
    monkeypatch.setattr(
        content_context,
        "current_context_binding",
        lambda: type("Binding", (), {"binding_id": ""})(),
    )

    with pytest.raises(content_model.ContentError) as excinfo:
        dispatch._materialize_invocation(
            "read",
            ["README.md"],
            content_model.CommonOptions(),
            b"hello\n",
            dirs=dirs,
        )

    assert "bind and launch a sibling actor" in str(excinfo.value)
    assert not (dirs.content_dir / "manifest.jsonl").exists()


def test_run_plugin_materializes_full_bytes_for_bounded_routed_read(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    assert session_plugin.main(["init", "--session", str(local_root)]) == 0
    capsys.readouterr()

    canonical = b'{"kind":"repo","title":"Title","lines":["line 1","line 2"]}'

    def fake_capture(argv: list[str], _options: object) -> Capture:
        return Capture(
            data=canonical,
            name="widgets.json",
            type="application/json",
        )

    def fake_project(argv: list[str], capture: Capture) -> bytes:
        assert capture.data == canonical
        return b"# Title\n\nline 1\nline 2\n"

    monkeypatch.setattr(
        read_plugin,
        "get_plugin",
        lambda name: (
            SimpleNamespace(capture=fake_capture, project=fake_project)
            if name == "github"
            else plugin_api.get_plugin(name)
        ),
    )

    assert (
        dispatch.run_plugin(
            "read",
            [
                "https://github.com/acme/widgets",
                "--head",
                "3",
                "--session",
                str(local_root),
            ],
        )
        == 0
    )
    captured = capsys.readouterr()
    assert captured.out == "# Title\n\nline 1\n"

    snapshots = content_store.scan_content_store(local_root / "content")
    assert len(snapshots) == 1
    snapshot = snapshots[0]
    assert snapshot.metadata["canonical_locator"] == "https://github.com/acme/widgets"
    assert snapshot.metadata["content_type"] == "application/json"
    assert snapshot.data_path.read_bytes() == canonical


def test_repeated_bounded_and_unbounded_read_share_one_canonical_snapshot(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    assert session_plugin.main(["init", "--session", str(local_root)]) == 0
    capsys.readouterr()

    canonical = b'{"kind":"repo","title":"Title","lines":["line 1","line 2"]}'

    def fake_capture(argv: list[str], _options: object) -> Capture:
        return Capture(
            data=canonical,
            name="widgets.json",
            type="application/json",
        )

    def fake_project(argv: list[str], capture: Capture) -> bytes:
        assert capture.data == canonical
        return b"# Title\n\nline 1\nline 2\n"

    monkeypatch.setattr(
        read_plugin,
        "get_plugin",
        lambda name: (
            SimpleNamespace(capture=fake_capture, project=fake_project)
            if name == "github"
            else plugin_api.get_plugin(name)
        ),
    )

    assert (
        dispatch.run_plugin(
            "read",
            [
                "https://github.com/acme/widgets",
                "--head",
                "3",
                "--session",
                str(local_root),
            ],
        )
        == 0
    )
    capsys.readouterr()
    assert (
        dispatch.run_plugin(
            "read", ["https://github.com/acme/widgets", "--session", str(local_root)]
        )
        == 0
    )
    capsys.readouterr()

    snapshots = content_store.scan_content_store(local_root / "content")
    assert len(snapshots) == 1
    assert (
        snapshots[0].metadata["canonical_locator"] == "https://github.com/acme/widgets"
    )

    assert (
        session_plugin.main(
            ["manifest", "--session", str(local_root), "--output", "json"]
        )
        == 0
    )
    payload = json.loads(capsys.readouterr().out)
    assert payload["entryCount"] == 1
    assert payload["fetchRecordCount"] == 2
    assert (
        payload["entries"][0]["canonical_locator"] == "https://github.com/acme/widgets"
    )
    assert payload["entries"][0]["fetchCount"] == 2


def test_github_route_prefers_markdown_for_common_github_targets() -> None:
    route_target = plugin_api.get_plugin("github").route_target
    assert route_target is not None

    assert route_target("https://github.com/acme/widgets") == [
        "https://github.com/acme/widgets"
    ]
    assert route_target("https://github.com/acme/widgets#readme") == [
        "https://github.com/acme/widgets#readme"
    ]
    assert route_target("https://github.com/acme/widgets/tree/main/docs") == [
        "https://github.com/acme/widgets/tree/main/docs"
    ]
    assert route_target("https://github.com/acme/widgets/tree/main/docs#readme") == [
        "https://github.com/acme/widgets/tree/main/docs#readme"
    ]
    assert route_target("https://github.com/acme/widgets/pull/42") == [
        "https://github.com/acme/widgets/pull/42"
    ]
    assert route_target("https://github.com/acme/widgets/releases/tag/v1.2.3") == [
        "https://github.com/acme/widgets/releases/tag/v1.2.3"
    ]
    assert route_target("https://github.com/acme/widgets/actions/runs/123456789") == [
        "https://github.com/acme/widgets/actions/runs/123456789"
    ]
    assert route_target(
        "https://github.com/acme/widgets/actions/runs/123456789/job/987654321"
    ) == ["https://github.com/acme/widgets/actions/runs/123456789/job/987654321"]
    assert route_target("github:github.com/acme/widgets/blob/main/README.md") == [
        "https://github.com/acme/widgets/blob/main/README.md"
    ]
    assert route_target(
        "github:github.com/acme/widgets/actions/runs/123456789/job/987654321"
    ) == ["https://github.com/acme/widgets/actions/runs/123456789/job/987654321"]
    assert route_target("github:search --type pr --repo acme/widgets ABC") == [
        "search",
        "--type",
        "pr",
        "--repo",
        "acme/widgets",
        "ABC",
    ]


def test_canonical_locator_routes_are_followable_through_read_contract() -> None:
    assert plugin_api.get_plugin("jira").route_target("jira:PROJ-3960") == [
        "get",
        "PROJ-3960",
    ]
    assert plugin_api.get_plugin("jira").route_target("jira:status") == ["status"]
    assert plugin_api.get_plugin("jira").route_target("jira:projects") == ["projects"]
    assert plugin_api.get_plugin("granola").route_target("granola:status") == ["status"]
    assert plugin_api.get_plugin("granola").route_target("granola:list --limit 5") == [
        "list",
        "--limit",
        "5",
    ]
    assert plugin_api.get_plugin("granola").route_target(
        "granola:list --sort created --order asc --offset 10 --limit 5"
    ) == [
        "list",
        "--sort",
        "created",
        "--order",
        "asc",
        "--offset",
        "10",
        "--limit",
        "5",
    ]
    assert plugin_api.get_plugin("granola").route_target(
        "granola:search Architecture"
    ) == [
        "search",
        "Architecture",
    ]
    assert plugin_api.get_plugin("granola").route_target(
        "granola:search --time-range last_30_days Architecture"
    ) == [
        "search",
        "--time-range",
        "last_30_days",
        "Architecture",
    ]
    assert plugin_api.get_plugin("granola").route_target(
        "granola:transcript 11111111-1111-1111-1111-111111111111"
    ) == ["transcript", "11111111-1111-1111-1111-111111111111"]
    assert plugin_api.get_plugin("granola").route_target(
        "granola:search-transcript --all latency"
    ) == ["search-transcript", "--all", "latency"]
    assert plugin_api.get_plugin("granola").route_target(
        "granola:11111111-1111-1111-1111-111111111111"
    ) == ["get", "11111111-1111-1111-1111-111111111111"]
    assert plugin_api.get_plugin("granola").route_target(
        "granola:get 'Weekly Review'"
    ) == [
        "get",
        "Weekly Review",
    ]
    assert plugin_api.get_plugin("grafana").route_target("grafana:status") == ["status"]
    assert plugin_api.get_plugin("grafana").route_target(
        "grafana:datasources --limit 25 --offset 25"
    ) == [
        "datasources",
        "--limit",
        "25",
        "--offset",
        "25",
    ]
    assert plugin_api.get_plugin("grafana").route_target(
        "grafana:search Architecture"
    ) == [
        "search",
        "Architecture",
    ]
    assert plugin_api.get_plugin("grafana").route_target(
        "grafana:get demo-dashboard-uid"
    ) == [
        "get",
        "demo-dashboard-uid",
    ]
    assert plugin_api.get_plugin("jira").route_target("jira:search Architecture") == [
        "search",
        "Architecture",
    ]
    assert plugin_api.get_plugin("jira").route_target(
        "jira:projects --limit 25 --offset 25"
    ) == [
        "projects",
        "--limit",
        "25",
        "--offset",
        "25",
    ]
    assert plugin_api.get_plugin("jira").route_target("jira:sprints --project OPS") == [
        "sprints",
        "--project",
        "OPS",
    ]
    assert plugin_api.get_plugin("jira").route_target(
        "jira:issue-types --project OPS"
    ) == [
        "issue-types",
        "--project",
        "OPS",
    ]
    assert plugin_api.get_plugin("jira").route_target(
        "jira:fields --project OPS --type 'Service Request'"
    ) == [
        "fields",
        "--project",
        "OPS",
        "--type",
        "Service Request",
    ]


def test_granola_route_target_rejects_invalid_locators_quietly(capsys) -> None:
    assert (
        plugin_api.get_plugin("granola").route_target(
            "granola:search --after not-a-date latency"
        )
        is None
    )
    assert capsys.readouterr().err == ""
    assert plugin_api.get_plugin("jira").route_target("jira:transitions OPS-42") == [
        "transitions",
        "OPS-42",
    ]
    assert plugin_api.get_plugin("jira").route_target(
        "jira:add-to-sprint OPS-42 --current --project OPS"
    ) == [
        "add-to-sprint",
        "OPS-42",
        "--current",
        "--project",
        "OPS",
    ]
    assert plugin_api.get_plugin("confluence").route_target("confluence:10101") == [
        "get",
        "10101",
    ]
    assert plugin_api.get_plugin("confluence").route_target(
        "confluence:search Architecture"
    ) == ["search", "Architecture"]
    assert plugin_api.get_plugin("gdocs").route_target("gdocs:doc-123") == [
        "get",
        "doc-123",
    ]
    assert plugin_api.get_plugin("gdocs").route_target(
        "https://docs.google.com/document/d/doc-123/edit#heading=h.demo"
    ) == ["get", "https://docs.google.com/document/d/doc-123/edit"]
    assert plugin_api.get_plugin("gdocs").route_target("gdocs:search Architecture") == [
        "search",
        "Architecture",
    ]
    assert plugin_api.get_plugin("gsheets").route_target(
        "https://docs.google.com/spreadsheets/d/sheet-123/edit#gid=0"
    ) == [
        "get",
        "https://docs.google.com/spreadsheets/d/sheet-123/edit",
    ]
    assert plugin_api.get_plugin("gsheets").route_target("gsheets:sheet-123") == [
        "get",
        "sheet-123",
    ]
    assert plugin_api.get_plugin("gsheets").route_target(
        "gsheets:search Architecture"
    ) == [
        "search",
        "Architecture",
    ]
    assert (
        plugin_api.get_plugin("gdrive").route_target(
            "https://docs.google.com/spreadsheets/d/sheet-123/edit#gid=0"
        )
        is None
    )
    assert plugin_api.get_plugin("slack").route_target(
        "https://example.slack.com/archives/C12345678/p1773085070240949"
        "?thread_ts=1773085070.240949#replies"
    ) == [
        "get",
        "https://example.slack.com/archives/C12345678/p1773085070240949"
        "?thread_ts=1773085070.240949",
    ]
    assert plugin_api.get_plugin("gdrive").route_target("gdrive:file-123") == [
        "get",
        "file-123",
    ]
    assert plugin_api.get_plugin("gdrive").route_target(
        "gdrive:search Architecture"
    ) == [
        "search",
        "Architecture",
    ]
    assert plugin_api.get_plugin("slack").route_target(
        "slack:thread:C12345678:1773085070240949"
    ) == ["get", "C12345678:1773085070.240949"]
    assert plugin_api.get_plugin("slack").route_target(
        "https://example.slack.com/docs/T12345678/F12345678#fragment"
    ) == ["get", "https://example.slack.com/docs/T12345678/F12345678"]
    assert plugin_api.get_plugin("slack").route_target(
        "slack:doc:T12345678:F12345678"
    ) == [
        "get",
        "slack:doc:T12345678:F12345678",
    ]
    assert plugin_api.get_plugin("slack").route_target("slack:search Architecture") == [
        "search",
        "Architecture",
    ]
    assert plugin_api.get_plugin("slack").route_target(
        "slack:search --workspace demo --source archive ABC"
    ) == ["search", "--workspace", "demo", "--source", "archive", "ABC"]
    assert plugin_api.get_plugin("slack").route_target("slack:workspace:demo") == [
        "status",
        "--workspace",
        "demo",
        "--output",
        "summary",
    ]
    assert plugin_api.get_plugin("slack").route_target("slack:demo") == [
        "status",
        "--workspace",
        "demo",
        "--output",
        "summary",
    ]



"""Shared payload constants and coercion helpers."""

from __future__ import annotations

import time

from gotta.compat import datetime


ACTOR_RUNNING_STATUS = {
    "starting",
    "active",
    "closing",
    "producing_evidence",
}

ACTOR_TERMINAL_STATUS = {
    "completed",
    "failed",
    "incomplete",
    "rejected",
    "signed_off",
}

ACTOR_STARTUP_GRACE_SECONDS = 30


def int_value(value: object, *, default: int = 0) -> int:
    try:
        return int(str(value or default))
    except ValueError:
        return default


def lifecycle_entries(value: object) -> list[dict[str, object]]:
    if not isinstance(value, list):
        return []
    return [dict(item) for item in value if isinstance(item, dict)]


def iso_age_seconds(value: object) -> float | None:
    raw = str(value or "").strip()
    if not raw:
        return None
    try:
        parsed = datetime.fromisoformat(raw.replace("Z", "+00:00"))
    except ValueError:
        return None
    return time.time() - parsed.timestamp()



"""Shared parsing and resolution for `gotta read` targets."""

from __future__ import annotations

from typing import Any

from gotta.content.model import CommonOptions
from gotta.resolve.model import ReadRequest, ReadTarget
from gotta.resolve.parse import build_parser, parse_args
from gotta.resolve.route import discover_plugin_route
from gotta.resolve.target import resolve_target

__all__ = [
    "ReadRequest",
    "ReadTarget",
    "build_parser",
    "parse_args",
    "discover_plugin_route",
    "resolve_read_target",
    "should_materialize",
    "canonical_locator",
    "preferred_name",
]


def resolve_read_target(
    argv: list[str],
    options: CommonOptions | Any | None = None,
) -> ReadTarget:
    return resolve_target(parse_args(argv), options or CommonOptions())


def should_materialize(argv: list[str]) -> bool:
    try:
        return resolve_read_target(argv).should_materialize
    except SystemExit:
        return False


def canonical_locator(argv: list[str]) -> str:
    try:
        return resolve_read_target(argv).canonical_locator
    except SystemExit:
        return "read"


def preferred_name(argv: list[str], options: CommonOptions | Any) -> str:
    try:
        return resolve_read_target(argv, options).preferred_name
    except SystemExit:
        save_as = str(getattr(options, "save_as", "") or "").strip()
        return save_as or "read.txt"



"""Actor launch blocker and rewrite detection helpers."""

from __future__ import annotations

from pathlib import Path

from gotta.session.bootstrap import (
    _bootstrap_actor_goal,
    _bootstrap_actor_want,
    _bootstrap_want,
)
from gotta.session.registry import (
    WANT_FILE,
    _actor_charter_command,
    _actor_dir_path,
    _actor_goal_path,
    _actor_label,
    _actor_session_dir,
    _actor_want_path,
    _resolve_bound_actor_name,
)


def _want_rewrite_pending(work_dir: Path) -> bool:
    want_path = work_dir / WANT_FILE
    if not want_path.is_file():
        return True
    current = want_path.read_text(encoding="utf-8").strip()
    return current == _bootstrap_want().strip()


def _goal_rewrite_pending(goal_path: Path) -> bool:
    if not goal_path.is_file():
        return True
    return (
        goal_path.read_text(encoding="utf-8")
        .strip()
        .startswith("# Seed Goal Placeholder")
    )


def _actor_want_rewrite_pending(work_root: Path, actor_name: str) -> bool:
    path = _actor_want_path(work_root, actor_name)
    if not path.is_file():
        return True
    current = path.read_text(encoding="utf-8").strip()
    return (
        current
        == _bootstrap_actor_want(
            actor_name=actor_name,
            label=_actor_label(actor_name, work_dir=work_root),
        ).strip()
    )


def _actor_goal_rewrite_pending(work_root: Path, actor_name: str) -> bool:
    path = _actor_goal_path(work_root, actor_name)
    if not path.is_file():
        return True
    return (
        path.read_text(encoding="utf-8").strip()
        == _bootstrap_actor_goal(
            actor_name=actor_name,
            label=_actor_label(actor_name, work_dir=work_root),
            actor_dir=_actor_session_dir(work_root, actor_name),
            work_dir=work_root,
        ).strip()
    )


def _actor_launch_blockers(work_root: Path, *, actor_name: str = "") -> list[str]:
    blockers: list[str] = []
    goal_path = work_root / "GOAL.md"
    if _want_rewrite_pending(work_root):
        blockers.append(f"rewrite `{work_root / WANT_FILE}` first")
    if _goal_rewrite_pending(goal_path):
        blockers.append(f"rewrite `{goal_path}` from the current moment before launch")
    if actor_name:
        actor_name = _resolve_bound_actor_name(work_root, actor_name)
        actor_want = _actor_dir_path(work_root, actor_name) / WANT_FILE
        actor_goal = _actor_dir_path(work_root, actor_name) / "GOAL.md"
        want_cmd = _actor_charter_command(actor_name, "want")
        goal_cmd = _actor_charter_command(actor_name, "goal")
        if _actor_want_rewrite_pending(work_root, actor_name):
            blockers.append(
                f"rewrite `{actor_want}` as the actor-local intent frame with `{want_cmd}` before launch"
            )
        if _actor_goal_rewrite_pending(work_root, actor_name):
            blockers.append(
                f"rewrite `{actor_goal}` as the actor-local goal with `{goal_cmd}` before launch"
            )
    return blockers



"""Low-level append helpers for session activity state."""

from __future__ import annotations

import json
import os
from pathlib import Path

from gotta.content.file import PRIVATE_FILE_MODE, ensure_private_dir


def _append_chunk(path: Path, chunk: str) -> None:
    ensure_private_dir(path.parent)
    fd = os.open(path, os.O_APPEND | os.O_CREAT | os.O_WRONLY, PRIVATE_FILE_MODE)
    try:
        os.write(fd, chunk.encode("utf-8"))
    finally:
        os.close(fd)
    try:
        path.chmod(PRIVATE_FILE_MODE)
    except OSError:
        pass


def _append_jsonl(path: Path, payload: dict[str, object]) -> None:
    _append_chunk(path, json.dumps(payload, sort_keys=True) + "\n")



from __future__ import annotations
import io
from pathlib import Path
import socket
from types import SimpleNamespace
import urllib.error

import pytest

import gotta.content.env as content_env
import gotta.content.model as content_model
import gotta.content.path as content_path
import gotta.content.store as content_store
from gotta.plugins import actor, read
from gotta.plugins.session import main as session
import gotta.resolve.read as resolve_read
from gotta.session import registry as session_registry


def test_read_local_directory_renders_native_listing(tmp_path: Path, capsys) -> None:
    root = tmp_path / "repo"
    root.mkdir()
    (root / ".git").mkdir()
    (root / ".git" / "config").write_text(
        "[core]\n\trepositoryformatversion = 0\n", encoding="utf-8"
    )
    (root / "docs").mkdir()
    (root / "docs" / "README.md").write_text("# Docs\n", encoding="utf-8")

    assert read.main([str(root), "--recursive", "--max-depth", "2"]) == 0

    output = capsys.readouterr().out
    assert f"# Directory: {root.resolve()}" in output
    assert "`.git/`" in output
    assert "`.git/config`" in output
    assert "`docs/`" in output
    assert "`docs/README.md`" in output


def test_read_requires_session_context_for_relative_hidden_file(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    repo = tmp_path / "repo"
    repo.mkdir()
    (repo / ".git").mkdir()
    config = repo / ".git" / "config"
    config.write_text(
        '[remote "origin"]\n\turl = git@github.com:acme/widgets.git\n', encoding="utf-8"
    )

    monkeypatch.chdir(tmp_path)
    monkeypatch.setenv("GOTTA_SESSION_REPO", str(repo))

    assert read.main([".git/config"]) == 2
    err = capsys.readouterr().err
    assert "requires an active or explicit session root" in err


def test_read_prefers_session_relative_target_over_shell_cwd(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    cwd_root = tmp_path / "cwd"
    session_root = tmp_path / "session"
    cwd_root.mkdir()
    session_root.mkdir()
    (cwd_root / "BRIEF.md").write_text("cwd copy\n", encoding="utf-8")
    (session_root / "BRIEF.md").write_text("session copy\n", encoding="utf-8")

    monkeypatch.chdir(cwd_root)
    monkeypatch.setenv("GOTTA_SESSION_DIR", str(session_root))

    assert read.main(["BRIEF.md"]) == 0
    output = capsys.readouterr().out
    assert "session copy" in output
    assert "cwd copy" not in output


def test_read_can_follow_stored_content_digest(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content_store.materialize_bytes(
        b"# Stored body\n\nline two\n",
        dirs=dirs,
        preferred_name="stored.md",
        metadata={"plugin": "read", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert read.main([result.digest]) == 0
    output = capsys.readouterr().out
    assert "# Stored body" in output


def test_read_can_follow_explicit_content_locator(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content_store.materialize_bytes(
        b"# Stored body\n\nline two\n",
        dirs=dirs,
        preferred_name="stored.md",
        metadata={"plugin": "read", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert read.main([f"content:{result.digest}"]) == 0
    output = capsys.readouterr().out
    assert "# Stored body" in output


def test_read_can_follow_explicit_content_locator_with_explicit_session(
    tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)
    result = content_store.materialize_bytes(
        b"# Stored body\n\nline two\n",
        dirs=dirs,
        preferred_name="stored.md",
        metadata={"plugin": "read", "locator": "demo", "canonical_locator": "demo"},
    )

    assert (
        read.main(["--session", str(dirs.session_dir), f"content:{result.digest}"]) == 0
    )
    output = capsys.readouterr().out
    assert "# Stored body" in output


def test_read_can_follow_unique_session_artifact_name(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_store.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert read.main(["slack-search-abc.md"]) == 0
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_can_follow_explicit_artifact_locator(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content_store.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert (
        read.main([content_path.artifact_locator("slack-search-abc.md", result.digest)])
        == 0
    )
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_can_follow_explicit_artifact_locator_with_explicit_session(
    tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)
    result = content_store.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    assert (
        read.main(
            [
                "--session",
                str(dirs.session_dir),
                content_path.artifact_locator("slack-search-abc.md", result.digest),
            ]
        )
        == 0
    )
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_can_follow_artifact_locator_from_session_root_state_env(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)
    result = content_store.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.delenv("GOTTA_SESSION_DIR", raising=False)
    monkeypatch.delenv("GOTTA_SESSION_CONTENT_DIR", raising=False)
    monkeypatch.chdir(dirs.session_dir)

    assert (
        read.main([content_path.artifact_locator("slack-search-abc.md", result.digest)])
        == 0
    )
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_can_follow_artifact_locator_from_actor_root_state_env(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content_env.write_state_env(dirs)
    result = content_store.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )
    actor_root = tmp_path / "session-actor"
    actor_state = actor_root / "state"
    actor_state.mkdir(parents=True, exist_ok=True)
    (actor_state / "env").write_text(
        "\n".join(
            [
                f"export GOTTA_SESSION_DIR='{actor_root}'",
                f"export GOTTA_SESSION_CONTENT_DIR='{dirs.content_dir}'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    monkeypatch.delenv("GOTTA_SESSION_DIR", raising=False)
    monkeypatch.delenv("GOTTA_SESSION_CONTENT_DIR", raising=False)
    monkeypatch.chdir(actor_root)

    assert (
        read.main([content_path.artifact_locator("slack-search-abc.md", result.digest)])
        == 0
    )
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_missing_actor_local_surface_reports_missing_local_path(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    actor_root = tmp_path / "session-claude"
    actor_root.mkdir()

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(actor_root))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(actor_root / "content"))

    assert read.main(["WANT.md"]) == 2
    err = capsys.readouterr().err
    assert "local target 'WANT.md' does not exist" in err
    assert "unsupported target" not in err


def test_read_missing_actor_goal_surface_reports_missing_local_path(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    actor_root = tmp_path / "session-claude"
    actor_root.mkdir()

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(actor_root))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(actor_root / "content"))

    assert read.main(["GOAL.md"]) == 2
    err = capsys.readouterr().err
    assert "local target 'GOAL.md' does not exist" in err
    assert "unsupported target" not in err


def test_read_seeded_actor_local_charter_surfaces(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session-root"

    assert session.main(["init", "--session", str(root)]) == 0
    capsys.readouterr()
    (root / "WANT.md").write_text("# Want\n\nReal intent.\n", encoding="utf-8")
    (root / "GOAL.md").write_text("# Goal\n\nReal goal.\n", encoding="utf-8")

    assert actor.main(["bind", "Claude", "--session", str(root)]) == 0
    capsys.readouterr()

    actor_root = session_registry._actor_session_dir(root, "claude")
    monkeypatch.setenv("GOTTA_SESSION_DIR", str(actor_root))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(actor_root / "content"))

    assert read.main(["WANT.md"]) == 0
    want_output = capsys.readouterr().out
    assert "# Actor Want Placeholder" in want_output

    assert read.main(["GOAL.md"]) == 0
    goal_output = capsys.readouterr().out
    assert "# Seed Actor Goal Placeholder" in goal_output


def test_read_does_not_materialize_local_artifact_rereads(
    monkeypatch, tmp_path: Path
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content_store.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert not resolve_read.should_materialize(
        [
            content_path.artifact_locator("slack-search-abc.md", result.digest),
            "--head",
            "2",
        ]
    )


def test_read_ambiguous_session_artifact_name_suggests_explicit_artifact_locators(
    monkeypatch, tmp_path: Path
) -> None:
    session_root = tmp_path / "session"
    dirs = content_model.ResolvedDirs(
        session_dir=session_root, content_dir=session_root / "content"
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    first = content_store.materialize_bytes(
        b"# Search Artifact A\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={
            "plugin": "slack",
            "locator": "demo-a",
            "canonical_locator": "demo-a",
        },
    )
    second = content_store.materialize_bytes(
        b"# Search Artifact B\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={
            "plugin": "slack",
            "locator": "demo-b",
            "canonical_locator": "demo-b",
        },
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    with pytest.raises(SystemExit) as excinfo:
        read.main(["slack-search-abc.md"])
    message = str(excinfo.value)
    assert "artifact:slack-search-abc.md@" in message
    assert content_path.artifact_locator("slack-search-abc.md", first.digest) in message
    assert (
        content_path.artifact_locator("slack-search-abc.md", second.digest) in message
    )


def test_read_supports_bounded_local_views(tmp_path: Path, capsys) -> None:
    target = tmp_path / "notes.md"
    target.write_text(
        "# Intro\n\nline 1\nline 2\n## Details\n\na\nb\nc\n",
        encoding="utf-8",
    )

    assert read.main([str(target), "--section", "Details", "--head", "2"]) == 0
    output = capsys.readouterr().out
    assert "## Details" in output
    assert "a" in output
    assert "b" not in output


def test_read_supports_bounded_provider_views(monkeypatch, capsys) -> None:
    monkeypatch.setattr(
        read,
        "resolve_read_target",
        lambda argv: resolve_read.ReadTarget(
            request=resolve_read.parse_args(argv),
            kind="routed",
            path=None,
            routed_plugin="github",
            routed_argv=["fake"],
            canonical_locator="https://github.com/acme/widgets",
            preferred_name="widgets.md",
            should_materialize=True,
        ),
    )

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        print("# Title")
        print("")
        print("line 1")
        print("line 2")
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["https://github.com/acme/widgets", "--head", "3"]) == 0
    output = capsys.readouterr().out
    assert "# Title" in output
    assert "line 1" in output
    assert "line 2" not in output


def test_read_fetch_url_summarizes_html_error_pages(monkeypatch) -> None:
    def fake_urlopen(_request, timeout=None):
        raise urllib.error.HTTPError(
            url="https://github.com/acme/widgets/blob/main/missing.md",
            code=404,
            msg="Not Found",
            hdrs={"Content-Type": "text/html; charset=utf-8"},
            fp=io.BytesIO(
                b"<html><head><title>Page not found \xc2\xb7 GitHub \xc2\xb7 GitHub</title></head><body>missing</body></html>"
            ),
        )

    monkeypatch.setattr(read.urllib.request, "urlopen", fake_urlopen)

    with pytest.raises(RuntimeError) as excinfo:
        read.fetch_url("https://github.com/acme/widgets/blob/main/missing.md")

    assert (
        str(excinfo.value)
        == "download failed with 404: Page not found · GitHub · GitHub"
    )


def test_read_fetch_url_uses_timeout_and_caps_remote_body(monkeypatch) -> None:
    seen_timeouts: list[float] = []

    class FakeResponse:
        def __init__(self) -> None:
            self.headers = {"Content-Type": "text/plain"}
            self._chunks = [
                b"a" * (read.REMOTE_FETCH_MAX_BYTES - 8),
                b"b" * 64,
            ]

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            if exc_type or exc or tb:
                return False
            return False

        def read(self, _size: int = -1) -> bytes:
            return self._chunks.pop(0) if self._chunks else b""

    def fake_urlopen(_request, timeout=None):
        seen_timeouts.append(timeout)
        return FakeResponse()

    monkeypatch.setattr(read.urllib.request, "urlopen", fake_urlopen)

    data, content_type, truncated = read.fetch_url("https://example.com/large")

    assert seen_timeouts == [read.REMOTE_FETCH_TIMEOUT_SECONDS]
    assert content_type == "text/plain"
    assert len(data) == read.REMOTE_FETCH_MAX_BYTES
    assert truncated is True


def test_read_fetch_url_reports_timeouts_cleanly(monkeypatch) -> None:
    def fake_urlopen(_request, timeout=None):
        raise socket.timeout("timed out")

    monkeypatch.setattr(read.urllib.request, "urlopen", fake_urlopen)

    with pytest.raises(RuntimeError) as excinfo:
        read.fetch_url("https://example.com/slow")

    assert str(excinfo.value) == "download timed out after 15s"


def test_read_response_body_does_not_mark_exact_cap_as_truncated() -> None:
    data, truncated = read._read_response_body(
        io.BytesIO(b"a" * read.REMOTE_FETCH_MAX_BYTES)
    )

    assert len(data) == read.REMOTE_FETCH_MAX_BYTES
    assert truncated is False


def test_read_remote_url_reports_truncation_and_still_applies_head(
    monkeypatch, capsys
) -> None:
    monkeypatch.setattr(
        read,
        "resolve_read_target",
        lambda argv: resolve_read.ReadTarget(
            request=resolve_read.parse_args(argv),
            kind="remote_url",
            path=None,
            routed_plugin=None,
            routed_argv=[],
            canonical_locator="https://example.com/large",
            preferred_name="large.txt",
            should_materialize=True,
        ),
    )
    monkeypatch.setattr(
        read,
        "fetch_url",
        lambda _target: (b"line 1\nline 2\n", "text/plain", True),
    )

    assert read.main(["https://example.com/large", "--head", "1"]) == 0
    captured = capsys.readouterr()
    assert captured.out == "line 1\n"
    assert "truncated remote body" in captured.err


def test_read_explicit_stdin_target_supports_bounded_view(monkeypatch, capsys) -> None:
    stdin = io.TextIOWrapper(io.BytesIO(b"line 1\nline 2\n"), encoding="utf-8")
    monkeypatch.setattr(read.sys, "stdin", stdin)

    assert read.main(["-", "--head", "1"]) == 0

    captured = capsys.readouterr()
    assert captured.out == "line 1\n"
    assert captured.err == ""


def test_read_view_shaping_makes_routed_targets_non_materializing() -> None:
    resolved = resolve_read.resolve_read_target(
        ["https://github.com/acme/widgets", "--head", "3"]
    )

    assert resolved.kind == "routed"
    assert resolved.should_materialize is True


def test_read_view_shaping_makes_remote_urls_non_materializing() -> None:
    resolved = resolve_read.resolve_read_target(
        ["https://example.com/manual.txt", "--tail", "5"]
    )

    assert resolved.kind == "remote_url"
    assert resolved.should_materialize is True


def test_read_whitespace_only_section_normalizes_to_plain_read() -> None:
    resolved = resolve_read.resolve_read_target(
        ["https://example.com/manual.txt", "--section", "   "]
    )

    assert resolved.kind == "remote_url"
    assert resolved.request.section == ""
    assert resolved.should_materialize is True


def test_read_help_text_describes_plain_vs_shaped_materialization() -> None:
    description = resolve_read.build_parser().description or ""

    assert (
        "Remote/provider reads store durable evidence only when an initialized session"
        in description
    )
    assert (
        "`--head`, `--tail`, and `--section` only trim what is shown to the operator"
        in read.USAGE
    )
    assert "--actor" in resolve_read.build_parser().format_help()


def test_execute_materializing_read_keeps_full_routed_bytes_under_bounded_view(
    monkeypatch,
) -> None:
    monkeypatch.setattr(
        read,
        "resolve_read_target",
        lambda argv: resolve_read.ReadTarget(
            request=resolve_read.parse_args(argv),
            kind="routed",
            path=None,
            routed_plugin="github",
            routed_argv=["fake"],
            canonical_locator="https://github.com/acme/widgets",
            preferred_name="widgets.json",
            should_materialize=True,
        ),
    )

    canonical = b'{"kind":"repo","title":"Title","lines":["line 1","line 2"]}'

    def fake_capture(argv: list[str], _options: object):
        return read.Capture(
            data=canonical,
            name="widgets.json",
            type="application/json",
        )

    def fake_project(argv: list[str], capture):
        assert capture.data == canonical
        return b"# Title\n\nline 1\nline 2\n"

    monkeypatch.setattr(
        read,
        "get_plugin",
        lambda name: (
            SimpleNamespace(capture=fake_capture, project=fake_project)
            if name == "github"
            else None
        ),
    )

    outcome = read.execute_materializing_read(
        ["https://github.com/acme/widgets", "--head", "3"]
    )

    assert outcome.code == 0
    assert outcome.display_bytes.decode("utf-8") == "# Title\n\nline 1\n"
    assert outcome.canonical_bytes == canonical


def test_execute_materializing_read_keeps_full_remote_bytes_under_bounded_view(
    monkeypatch,
) -> None:
    monkeypatch.setattr(
        read,
        "fetch_url",
        lambda _target: (b"line 1\nline 2\nline 3\n", "text/plain", False),
    )

    outcome = read.execute_materializing_read(
        ["https://example.com/manual.txt", "--tail", "2"]
    )

    assert outcome.code == 0
    assert outcome.display_bytes.decode("utf-8") == "line 2\nline 3\n"
    assert outcome.canonical_bytes.decode("utf-8") == "line 1\nline 2\nline 3\n"


def test_read_passes_provider_flags_through_for_routed_targets(monkeypatch) -> None:
    resolved = resolve_read.resolve_read_target(
        ["--limit", "10", "https://github.com/acme/widgets/commits/main"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "github"
    assert resolved.routed_argv == [
        "--limit",
        "10",
        "https://github.com/acme/widgets/commits/main",
    ]


def test_read_passes_provider_flags_through_after_routed_target(monkeypatch) -> None:
    resolved = resolve_read.resolve_read_target(
        ["https://github.com/acme/widgets/commits/main", "--limit", "10"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "github"
    assert resolved.routed_argv == [
        "https://github.com/acme/widgets/commits/main",
        "--limit",
        "10",
    ]


def test_read_uses_provider_normalized_fragmentless_url_for_routed_targets() -> None:
    resolved = resolve_read.resolve_read_target(
        ["https://docs.google.com/spreadsheets/d/sheet-123/edit#gid=0"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "gsheets"
    assert resolved.routed_argv == [
        "get",
        "https://docs.google.com/spreadsheets/d/sheet-123/edit",
    ]


def test_read_preserves_github_fragment_hint_while_canonicalizing_identity() -> None:
    resolved = resolve_read.resolve_read_target(
        ["https://github.com/acme/widgets#readme"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "github"
    assert resolved.routed_argv == ["https://github.com/acme/widgets#readme"]
    assert resolved.canonical_locator == "https://github.com/acme/widgets"


def test_read_routes_github_actions_job_urls() -> None:
    resolved = resolve_read.resolve_read_target(
        ["https://github.com/acme/widgets/actions/runs/123456789/job/987654321"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "github"
    assert resolved.routed_argv == [
        "https://github.com/acme/widgets/actions/runs/123456789/job/987654321"
    ]
    assert (
        resolved.canonical_locator
        == "https://github.com/acme/widgets/actions/runs/123456789/job/987654321"
    )


def test_read_delegates_github_commit_history_limit_after_target(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert (
        read.main(["https://github.com/acme/widgets/commits/main", "--limit", "10"])
        == 0
    )
    assert seen == [
        (
            "github",
            ["https://github.com/acme/widgets/commits/main", "--limit", "10"],
        )
    ]


def test_read_can_follow_canonical_jira_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:PROJ-3960"]) == 0
    assert seen == [("jira", ["get", "PROJ-3960"])]


def test_read_can_follow_canonical_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:search Architecture"]) == 0
    assert seen == [("jira", ["search", "Architecture"])]


def test_read_can_follow_canonical_granola_document_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:11111111-1111-1111-1111-111111111111"]) == 0
    assert seen == [("granola", ["get", "11111111-1111-1111-1111-111111111111"])]


def test_read_can_follow_canonical_granola_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:search --limit 5 latency"]) == 0
    assert seen == [("granola", ["search", "--limit", "5", "latency"])]


def test_read_can_follow_canonical_granola_transcript_search_locator(
    monkeypatch,
) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:search-transcript --all latency"]) == 0
    assert seen == [("granola", ["search-transcript", "--all", "latency"])]


def test_read_can_follow_canonical_granola_transcript_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:transcript 11111111-1111-1111-1111-111111111111"]) == 0
    assert seen == [("granola", ["transcript", "11111111-1111-1111-1111-111111111111"])]


def test_read_can_follow_canonical_search_locator_with_flags(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert (
        read.main(
            [
                "slack:search --workspace example-workspace --source archive --limit 10 --output markdown ABC"
            ]
        )
        == 0
    )
    assert seen == [
        (
            "slack",
            [
                "search",
                "--workspace",
                "example-workspace",
                "--source",
                "archive",
                "--limit",
                "10",
                "--output",
                "markdown",
                "ABC",
            ],
        )
    ]


def test_read_can_follow_canonical_search_locator_with_equals_flags_unquoted(
    monkeypatch,
) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:search", "--limit=10", "Architecture"]) == 0
    assert seen == [("jira", ["search", "--limit", "10", "Architecture"])]


def test_read_can_follow_canonical_search_locator_with_flags_after_query(
    monkeypatch,
) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert (
        read.main(["slack:search ABC reboot --workspace example-workspace --limit 10"])
        == 0
    )
    assert seen == [
        (
            "slack",
            [
                "search",
                "--workspace",
                "example-workspace",
                "--limit",
                "10",
                "ABC reboot",
            ],
        )
    ]


def test_read_can_follow_canonical_github_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["github:search --type pr --repo acme/widgets ABC"]) == 0
    assert seen == [
        ("github", ["search", "--type", "pr", "--repo", "acme/widgets", "ABC"])
    ]


def test_read_can_follow_canonical_global_github_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["github:search --global ABC"]) == 0
    assert seen == [("github", ["search", "--global", "ABC"])]


def test_read_can_follow_slack_workspace_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["slack:example-workspace"]) == 0
    assert seen == [
        ("slack", ["status", "--workspace", "example-workspace", "--output", "summary"])
    ]


def test_read_joins_unquoted_multiword_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:search", "ABC", "reboot"]) == 0
    assert seen == [("jira", ["search", "ABC reboot"])]


def test_read_accepts_option_like_locator_tails_after_end_of_options(
    monkeypatch,
) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["--", "jira:search", "ABC", "--head", "180"]) == 0
    assert seen == [("jira", ["search", "ABC --head 180"])]



"""Read paths for `gotta notes`."""

from __future__ import annotations

import argparse
import json

from gotta.notes.render import actor_notes_payload, render_actor_notes_markdown
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope
from gotta.session.activity.note import _record_note_check
from gotta.session.status.payload.main import _actor_status_payload

from .render import render_session_text


def _entry_records(value: object) -> list[dict[str, object]]:
    if not isinstance(value, list):
        return []
    return [dict(item) for item in value if isinstance(item, dict)]


def show_notes(args: argparse.Namespace) -> int:
    work_dir, scoped_actor = session_scope._observation_scope(
        explicit_session=getattr(args, "session", None),
        explicit_actor=getattr(args, "actor", None),
    )
    if scoped_actor:
        _record_note_check(work_dir, scoped_actor)
        status = _actor_status_payload(work_dir, scoped_actor)
        payload = actor_notes_payload(
            work_dir,
            scoped_actor,
            label=session_registry._actor_label(scoped_actor, work_dir=work_dir),
            status_payload=status,
        )
        if args.output == "json":
            print(json.dumps(payload, indent=2, sort_keys=True))
            return 0
        print(
            render_actor_notes_markdown(
                work_dir,
                scoped_actor,
                label=session_registry._actor_label(scoped_actor, work_dir=work_dir),
                status_payload=status,
            ),
            end="",
        )
        return 0
    actor_ids = list(session_scope._target_actor_ids(work_dir))
    if not actor_ids:
        raise SystemExit(
            "no actors bound for this session; bind one intentionally with "
            + session_registry._actor_bind_examples(prefix="gotta actor bind")
        )
    actor_payloads: dict[str, dict[str, object]] = {}
    entries: list[dict[str, object]] = []
    for actor_name in actor_ids:
        status = _actor_status_payload(work_dir, actor_name)
        actor_payload = actor_notes_payload(
            work_dir,
            actor_name,
            label=session_registry._actor_label(actor_name, work_dir=work_dir),
            status_payload=status,
        )
        actor_payloads[actor_name] = actor_payload
        for record in _entry_records(actor_payload.get("entries")):
            payload = dict(record)
            payload.setdefault("actor", actor_name)
            payload.setdefault(
                "label",
                session_registry._actor_label(actor_name, work_dir=work_dir),
            )
            entries.append(payload)
    entries.sort(key=lambda item: str(item.get("timestamp") or ""))
    payload = {
        "session_root": str(work_dir),
        "actor_count": len(actor_ids),
        "actors": actor_payloads,
        "entry_count": len(entries),
        "entries": entries,
    }
    if args.output == "json":
        print(json.dumps(payload, indent=2, sort_keys=True))
        return 0
    print(
        render_session_text(
            actor_count=len(actor_ids),
            entry_count=len(entries),
            entries=entries,
        )
    )
    return 0



"""Argument parsing for `gotta logs`."""

from __future__ import annotations

import argparse

from gotta.session import charter as session_charter


def _add_root_args(parser: argparse.ArgumentParser) -> None:
    session_charter.add_target_args(parser)


def build_parser(command_name: str = "gotta logs") -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=command_name,
        description=(
            "Inspect or mutate the canonical session procedural trace. "
            "Shared-session read paths show all bound actor logs by default. "
            "Use notes for actor-authored narration; use logs for chronology and "
            "system/runtime trace."
        ),
    )
    parser.add_argument(
        "action",
        nargs="?",
        choices=["show", "append", "extend"],
        help="show entries, append one entry, or extend with many entries",
    )
    parser.add_argument(
        "value",
        nargs="*",
        help="inline log text for append/extend",
    )
    _add_root_args(parser)
    parser.add_argument(
        "--from-file",
        help="read log text from a UTF-8 file instead of inline text; use '-' for stdin",
    )
    parser.add_argument(
        "--stdin",
        dest="use_stdin",
        action="store_true",
        help="read log text from stdin explicitly",
    )
    parser.add_argument("--output", choices=["json", "text"], default="text")
    parser.add_argument("--limit", type=int, default=0)
    return parser


def session_access_mode(argv: list[str]) -> str:
    positionals = session_charter.argv_positionals(
        argv,
        valued_flags=(
            "--session",
            "--actor",
            "--from-file",
            "--output",
            "--limit",
        ),
    )
    action = positionals[0] if positionals else "show"
    return "write" if action in {"append", "extend"} else "read"



"""Manifest payload synthesis."""

from __future__ import annotations

import re
from typing import Mapping

from gotta.content.model import ResolvedDirs
from gotta.content.path import content_locator

from ..core import (
    artifact_human_locator,
    artifact_kind,
    artifact_kind_counts,
    compile_filter_pattern,
    follow_command,
    match_any,
    match_filter_text,
    paginate_items,
    resolved_visibility_metadata,
    session_read_command,
    top_count_records,
)
from .record import (
    ManifestEntry,
    aggregate_manifest_entries,
    filter_manifest_entries,
    manifest_entries,
    manifest_entry_sort_key,
)


def _string(value: object) -> str:
    return str(value or "").strip()


def _string_list(value: object) -> list[str]:
    if not isinstance(value, list):
        return []
    values: list[str] = []
    for item in value:
        text = _string(item)
        if text:
            values.append(text)
    return values


def _manifest_entry_matches(
    entry: ManifestEntry,
    pattern: re.Pattern[str] | None,
) -> bool:
    return match_any(
        pattern,
        entry.get("canonical_locator"),
        entry.get("locator"),
        entry.get("preferred_name"),
        entry.get("plugin"),
        entry.get("plugins"),
        entry.get("actor"),
        entry.get("actors"),
        entry.get("locators"),
        entry.get("artifactKinds"),
        entry.get("artifactKind"),
        entry.get("artifact_locator"),
        entry.get("content_locator"),
        entry.get("visibility_basis"),
    )


def _rendered_manifest_entry(
    entry: Mapping[str, object],
    *,
    session_ref: str,
) -> ManifestEntry:
    checksum = _string(entry.get("checksum"))
    preferred_name = _string(entry.get("preferred_name")) or "data"
    canonical_locator = _string(entry.get("canonical_locator")) or _string(
        entry.get("locator")
    )
    return {
        **entry,
        "artifactKind": artifact_kind(entry.get("artifact_kind")),
        "content_locator": content_locator(checksum) if checksum else "",
        "fetchCount": int(_string(entry.get("fetchCount")) or 0),
        "firstFetchedAt": _string(entry.get("firstFetchedAt")),
        "lastFetchedAt": _string(entry.get("lastFetchedAt")),
        "plugins": _string_list(entry.get("plugins")),
        "actors": _string_list(entry.get("actors")),
        "locators": _string_list(entry.get("locators")),
        "artifactKinds": _string_list(entry.get("artifactKinds")),
        "artifact_locator": artifact_human_locator(preferred_name, checksum),
        "follow_command": follow_command(
            canonical_locator,
            checksum=checksum,
            session_ref=session_ref,
        ),
        "content_follow_command": (
            follow_command("", checksum=checksum, session_ref=session_ref)
            if checksum
            else ""
        ),
        "artifact_follow_command": (
            session_read_command(
                artifact_human_locator(preferred_name, checksum),
                session_ref=session_ref,
            )
            if checksum
            else ""
        ),
        **resolved_visibility_metadata(
            dict(entry),
            provider=_string(entry.get("plugin")),
            plugin=_string(entry.get("plugin")),
            subcommand=_string(entry.get("subcommand")),
            locator=canonical_locator,
        ),
    }


def manifest_payload(
    dirs: ResolvedDirs,
    *,
    plugin: str = "",
    actor: str = "",
    locator: str = "",
    filter_query: str = "",
    limit: int = 20,
    offset: int = 0,
    include_all: bool = False,
    session_ref: str = "",
) -> dict[str, object]:
    raw_entries = filter_manifest_entries(
        manifest_entries(dirs),
        plugin=plugin,
        actor=actor,
        locator=locator,
    )
    entries = aggregate_manifest_entries(raw_entries)
    filter_text = match_filter_text(filter_query)
    filter_pattern = compile_filter_pattern(filter_text)
    if filter_pattern is not None:
        entries = [
            entry for entry in entries if _manifest_entry_matches(entry, filter_pattern)
        ]
    ordered = sorted(entries, key=manifest_entry_sort_key, reverse=True)
    discovery_count, evidence_count = artifact_kind_counts(ordered)
    paged, paging = paginate_items(
        ordered,
        limit=limit,
        offset=offset,
        include_all=include_all,
    )
    rendered_entries = [
        _rendered_manifest_entry(
            {str(key): value for key, value in entry.items()},
            session_ref=session_ref,
        )
        for entry in paged
    ]
    fetch_record_count = sum(
        int(_string(entry.get("fetchCount")) or 0) for entry in ordered
    )
    top_plugins = top_count_records(
        [
            plugin_name
            for entry in ordered
            for plugin_name in _string_list(entry.get("plugins"))
        ],
        key="plugin",
    )
    top_actors = top_count_records(
        [
            actor_name
            for entry in ordered
            for actor_name in _string_list(entry.get("actors"))
        ],
        key="actor",
    )
    return {
        "sessionDir": str(dirs.session_dir),
        "contentDir": str(dirs.content_dir),
        "manifestPath": str(dirs.content_dir / "manifest.jsonl"),
        "sessionRef": session_ref,
        "entryCount": len(entries),
        "fetchRecordCount": fetch_record_count,
        **paging,
        "discoveryArtifactCount": discovery_count,
        "evidenceArtifactCount": evidence_count,
        "topPlugins": top_plugins,
        "topActors": top_actors,
        "pluginFilter": plugin,
        "actorFilter": actor,
        "locatorFilter": locator,
        "filter": filter_text,
        "entries": rendered_entries,
    }



"""Canonical capture objects for artifact-bearing retrieval."""

from __future__ import annotations

import argparse
from dataclasses import dataclass, field
import json
from typing import Any, Callable


@dataclass(frozen=True, slots=True)
class Capture:
    data: bytes
    name: str = ""
    type: str = ""
    meta: dict[str, Any] = field(default_factory=dict)
    view: dict[str, Any] = field(default_factory=dict)


def json_bytes(payload: Any) -> bytes:
    return (json.dumps(payload, indent=2, sort_keys=True) + "\n").encode("utf-8")


def capture_json_command(
    args: argparse.Namespace,
    runner: Callable[[argparse.Namespace], int],
    *,
    detail: str,
) -> bytes:
    from gotta.dispatch.stream import capture_stdout

    captured_args = argparse.Namespace(**vars(args))
    setattr(captured_args, "output", "json")
    with capture_stdout() as captured:
        code = runner(captured_args)
    if code != 0:
        message = captured.getvalue().decode("utf-8", errors="replace").strip()
        raise RuntimeError(message or detail)
    return captured.getvalue()



"""Next-step synthesis for actor status payloads."""

from __future__ import annotations

from gotta.session.status.payload.active import active_next_step
from gotta.session.status.payload.activity import note_check_next_step
from gotta.session.status.payload.closing import closing_next_step
from gotta.session.status.payload.pending import pending_next_step
from gotta.session.registry import _normalize_actor_name
from gotta.session.status.payload.value import int_value
from gotta.session.status.payload.runtime import low_signal_next_step, runtime_next_step


def actor_next_step(actor_name: str, payload: dict[str, object]) -> str:
    normalized_actor = _normalize_actor_name(actor_name)
    derived_status = str(payload.get("status") or "")
    next_step = runtime_next_step(normalized_actor, payload)
    if not next_step:
        next_step = closing_next_step(normalized_actor, payload)
    if not next_step:
        next_step = active_next_step(normalized_actor, payload)
    if not next_step:
        next_step = pending_next_step(normalized_actor, payload)
    low_signal_step = low_signal_next_step(payload)
    if low_signal_step:
        next_step = low_signal_step
    pulse_next_step = note_check_next_step(
        note_checks_since_update=int_value(payload.get("note_checks_since_update")),
        last_note_at=str(payload.get("last_note_at") or ""),
        derived_status=derived_status,
    )
    if pulse_next_step:
        next_step = f"{pulse_next_step} {next_step}".strip()
    return next_step



"""Mutation paths for `gotta notes`."""

from __future__ import annotations

import argparse
from pathlib import Path

from gotta.actor import require_writer, session_actor, writer_name
from gotta.content.context import current_actor
from gotta.notes.file import append_actor_note
from gotta.session import bootstrap as session_bootstrap
from gotta.session import charter as session_charter
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope
from gotta.session.activity.note import _reset_note_check_feedback
from gotta.session.activity.record import (
    _actor_log_line,
    _append_actor_event,
    _record_actor_surface_activity,
)


def _require_bound_actor(work_dir: Path, actor_name: str) -> None:
    if actor_name not in session_scope._selected_actor_ids(work_dir):
        raise SystemExit(
            f"{actor_name} is not bound for this session; bind them first with "
            f"`gotta actor bind {session_registry._actor_label(actor_name, work_dir=work_dir)}`"
        )


def _append_actor_name(work_dir: Path, *, explicit_actor: str = "") -> str:
    if explicit_actor:
        actor_name = session_registry._resolve_bound_actor_name(
            work_dir, explicit_actor
        )
        _require_bound_actor(work_dir, actor_name)
        return actor_name
    actor_ids = list(session_scope._target_actor_ids(work_dir))
    if not actor_ids:
        raise SystemExit(
            "no actor is in scope here; bind or enter one first, or pass "
            "`--actor <actor>` explicitly"
        )
    rooted_actor = session_actor(work_dir)
    if rooted_actor and rooted_actor in actor_ids:
        return rooted_actor
    ambient_actor = current_actor(default_actor="")
    if ambient_actor and ambient_actor in actor_ids:
        return ambient_actor
    if len(actor_ids) > 1:
        raise SystemExit(
            "actor note append is ambiguous across multiple bound actors; pass "
            "`--actor <actor>` explicitly"
        )
    return actor_ids[0]


def _append_author_name(
    actor_name: str, *, explicit_actor: str = "", writer: str = ""
) -> str:
    if explicit_actor:
        return writer or actor_name
    return actor_name


def append_note(args: argparse.Namespace) -> int:
    session_dir = session_scope._session_dir(
        explicit_session=getattr(args, "session", None),
        explicit_actor=getattr(args, "actor", None),
    )
    work_dir = session_dir.resolve()
    actor_name = _append_actor_name(
        work_dir,
        explicit_actor=str(getattr(args, "actor", None) or ""),
    )
    actor_writer = writer_name()
    author_name = _append_author_name(
        actor_name,
        explicit_actor=str(getattr(args, "actor", None) or ""),
        writer=actor_writer,
    )
    require_writer(
        work_dir,
        actor_name,
        writer=actor_writer,
        action="write into this actor branch",
    )
    session_bootstrap._ensure_actor_surface(work_dir, actor_name)
    message = session_charter._normalize_entry_text(
        session_charter._read_text_source(
            session_root=session_dir,
            inline=(args.value[0] if args.value else None),
            from_file=args.from_file,
            use_stdin=args.use_stdin,
            input_name="actor note",
        ),
        input_name="actor note",
    )
    append_actor_note(work_dir, actor_name, message=message, author=author_name)
    _reset_note_check_feedback(work_dir, actor_name)
    _append_actor_event(
        work_dir,
        actor_name,
        event="note",
        detail=message.splitlines()[0],
        author=author_name,
    )
    _actor_log_line(
        work_dir,
        actor_name,
        f"noted: {message.splitlines()[0]}",
        author=author_name,
    )
    _record_actor_surface_activity(
        work_dir,
        actor_name=actor_name,
        surface="notes",
        action="append",
        detail="appended actor note",
        actor=author_name,
    )
    print(f"appended actor note for {actor_name}")
    return 0



"""Argument parsing for top-level `gotta read` resolution."""

from __future__ import annotations

import argparse

from gotta.resolve.model import ReadRequest
from gotta.resolve.route import partition_routed_target_tokens


def _int_value(value: object, *, default: int) -> int:
    return value if isinstance(value, int) else default


def _string_value(value: object) -> str:
    return str(value or "").strip()


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="gotta read",
        description=(
            "Render one local or remote target through the native retrieval surface. "
            "Remote/provider reads store durable evidence only when an "
            "initialized session is already in play or passed explicitly; "
            "`--head`, `--tail`, and `--section` only change what is shown to "
            "the operator, while local/session-owned rereads stay as "
            "non-materializing views."
        ),
    )
    parser.add_argument("target", nargs="?")
    parser.add_argument("--session", help=argparse.SUPPRESS)
    parser.add_argument(
        "--actor",
        help="attribute any materialized artifact from this read to the selected bound actor",
    )
    parser.add_argument(
        "--recursive",
        action="store_true",
        help="traverse local directories recursively instead of listing only one level",
    )
    parser.add_argument(
        "--max-depth",
        type=int,
        default=3,
        help="maximum directory traversal depth when rendering local directories",
    )
    parser.add_argument(
        "--head", type=int, default=0, help="show only the first N lines"
    )
    parser.add_argument(
        "--tail", type=int, default=0, help="show only the last N lines"
    )
    parser.add_argument(
        "--section",
        help="show only the markdown section whose heading contains this text",
    )
    return parser


def parse_args(argv: list[str]) -> ReadRequest:
    parser = build_parser()
    if any(token in {"-h", "--help"} for token in argv):
        return ReadRequest(
            target=None,
            recursive=False,
            max_depth=3,
            head=0,
            tail=0,
            section="",
            session="",
            actor="",
            routed_plugin=None,
            routed_argv=(),
        )
    values: dict[str, object] = {
        "recursive": False,
        "max_depth": 3,
        "head": 0,
        "tail": 0,
        "section": "",
    }
    flags = {"--recursive": "recursive"}
    int_fields = {"--max-depth": "max_depth", "--head": "head", "--tail": "tail"}
    str_fields = {
        "--section": "section",
        "--session": "session",
        "--actor": "actor",
    }
    residual: list[str] = []
    index = 0
    while index < len(argv):
        token = argv[index]
        if token == "--":
            residual.extend(argv[index + 1 :])
            break
        name, has_inline, inline_value = token.partition("=")
        if name in flags and not has_inline:
            values[flags[name]] = True
            index += 1
            continue
        if name in int_fields or name in str_fields:
            if not has_inline:
                if index + 1 >= len(argv):
                    parser.error(f"argument {name}: expected one argument")
                inline_value = argv[index + 1]
                index += 1
            field = int_fields.get(name) or str_fields.get(name) or ""
            try:
                values[field] = (
                    int(inline_value) if name in int_fields else inline_value
                )
            except ValueError as exc:
                raise SystemExit(f"argument {name}: invalid int value") from exc
            index += 1
            continue
        residual.append(token)
        index += 1
    target: str | None = None
    routed_plugin: str | None = None
    routed_argv: tuple[str, ...] = ()
    routed = partition_routed_target_tokens(residual)
    if routed is not None:
        routed_plugin, target, routed_argv = routed
    elif residual:
        flagged = next(
            (token for token in residual if token != "-" and token.startswith("-")),
            "",
        )
        if flagged:
            parser.error(f"unrecognized arguments: {flagged}")
        target = (
            " ".join(part.strip() for part in residual if part.strip()).strip() or None
        )
    return ReadRequest(
        target=target,
        recursive=bool(values["recursive"]),
        max_depth=_int_value(values["max_depth"], default=3),
        head=_int_value(values["head"], default=0),
        tail=_int_value(values["tail"], default=0),
        section=_string_value(values["section"]),
        session=_string_value(values.get("session")),
        actor=_string_value(values.get("actor")),
        routed_plugin=routed_plugin,
        routed_argv=routed_argv,
    )



"""Read paths for `gotta logs`."""

from __future__ import annotations

import argparse
import json
from pathlib import Path

from gotta.actor import writer_role
from gotta.content.scope import session_is_initialized
from gotta.logs import log_records, logs_state_path
from gotta.session import charter as session_charter
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope
from gotta import topology

from .render import render_follow_text, render_session_text


def _int_value(value: object) -> int:
    return value if isinstance(value, int) else 0


def _entry_records(value: object) -> list[dict[str, object]]:
    if not isinstance(value, list):
        return []
    return [dict(item) for item in value if isinstance(item, dict)]


def _limited_records(
    records: list[dict[str, object]],
    *,
    limit: int,
) -> list[dict[str, object]]:
    bounded = max(limit, 0)
    return records[-bounded:] if bounded > 0 else records


def _aggregate_logs(
    work_dir: Path,
    *,
    actor_name: str | None = None,
) -> tuple[list[str], list[dict[str, object]]]:
    actor_ids = list(session_scope._target_actor_ids(work_dir, actor_name))
    records: list[dict[str, object]] = []
    for current_actor in actor_ids:
        actor_root = session_registry._actor_session_dir(work_dir, current_actor)
        for record in log_records(actor_root):
            if (
                actor_root.resolve().parent.name == "actors"
                and writer_role(
                    actor_root, current_actor, writer=str(record.get("actor") or "")
                )
                == "foreign"
            ):
                continue
            payload = dict(record)
            payload.setdefault("actor", current_actor)
            payload.setdefault(
                "label", session_registry._actor_label(current_actor, work_dir=work_dir)
            )
            records.append(payload)
    records.sort(key=lambda item: str(item.get("timestamp") or ""))
    return actor_ids, records


def _is_exact_session_root(work_dir: Path) -> bool:
    resolved = work_dir.resolve()
    return (
        resolved.parent.name != "actors"
        and topology.parse_grouped_session_root(resolved) is None
        and topology.parse_shared_session_root(resolved) is None
        and session_is_initialized(resolved)
    )


def _actor_payload(
    work_dir: Path,
    *,
    actor_name: str,
    limit: int,
) -> dict[str, object]:
    visible = [
        record
        for record in log_records(work_dir)
        if writer_role(work_dir, actor_name, writer=str(record.get("actor") or ""))
        != "foreign"
    ]
    entries = sorted(visible, key=lambda item: str(item.get("timestamp") or ""))
    return {
        "state_path": str(logs_state_path(work_dir)),
        "locator": session_charter._native_surface_locator(
            "logs", actor_name=actor_name
        ),
        "follow_command": session_charter._native_surface_follow_command(
            "logs",
            actor_name=actor_name,
        ),
        "entry_count": len(entries),
        "entries": _limited_records(entries, limit=limit),
    }


def _session_payload(work_dir: Path, *, limit: int) -> dict[str, object]:
    entries = sorted(
        log_records(work_dir), key=lambda item: str(item.get("timestamp") or "")
    )
    return {
        "session_root": str(work_dir),
        "actor_count": 0,
        "actors": [],
        "entry_count": len(entries),
        "entries": _limited_records(entries, limit=limit),
        "state_path": str(logs_state_path(work_dir)),
        "locator": session_charter._native_surface_locator("logs"),
        "follow_command": session_charter._native_surface_follow_command("logs"),
    }


def _aggregate_payload(
    work_dir: Path,
    *,
    actor_ids: list[str],
    records: list[dict[str, object]],
    limit: int,
) -> dict[str, object]:
    return {
        "session_root": str(work_dir),
        "actor_count": len(actor_ids),
        "actors": actor_ids,
        "entry_count": len(records),
        "entries": _limited_records(records, limit=limit),
        "state_paths": {
            actor: str(
                logs_state_path(session_registry._actor_session_dir(work_dir, actor))
            )
            for actor in actor_ids
        },
        "locators": {
            actor: session_charter._native_surface_locator("logs", actor_name=actor)
            for actor in actor_ids
        },
        "follow_commands": {
            actor: session_charter._native_surface_follow_command(
                "logs", actor_name=actor
            )
            for actor in actor_ids
        },
    }


def show_logs(args: argparse.Namespace) -> int:
    work_dir, scoped_actor = session_scope._observation_scope(
        explicit_session=getattr(args, "session", None),
        explicit_actor=getattr(args, "actor", None),
    )
    if scoped_actor:
        payload = _actor_payload(work_dir, actor_name=scoped_actor, limit=args.limit)
        if args.output == "json":
            print(json.dumps(payload, indent=2, sort_keys=True))
            return 0
        print(
            render_follow_text(
                follow_command=str(payload["follow_command"]),
                entry_count=_int_value(payload.get("entry_count")),
                entries=_entry_records(payload.get("entries")),
                default_actor=scoped_actor,
            )
        )
        return 0
    actor_ids, records = _aggregate_logs(work_dir)
    if not actor_ids:
        if not _is_exact_session_root(work_dir):
            raise SystemExit(
                "no actors bound for this session; bind one intentionally with "
                + session_registry._actor_bind_examples(prefix="gotta actor bind")
            )
        payload = _session_payload(work_dir, limit=args.limit)
        if args.output == "json":
            print(json.dumps(payload, indent=2, sort_keys=True))
            return 0
        print(
            render_follow_text(
                follow_command=str(payload["follow_command"]),
                entry_count=_int_value(payload.get("entry_count")),
                entries=_entry_records(payload.get("entries")),
                default_actor="session",
            )
        )
        return 0
    payload = _aggregate_payload(
        work_dir,
        actor_ids=actor_ids,
        records=records,
        limit=args.limit,
    )
    if args.output == "json":
        print(json.dumps(payload, indent=2, sort_keys=True))
        return 0
    print(
        render_session_text(
            actor_count=_int_value(payload.get("actor_count")),
            entry_count=_int_value(payload.get("entry_count")),
            entries=_entry_records(payload.get("entries")),
        )
    )
    return 0



"""Session and actor scope resolution helpers."""

from __future__ import annotations

import os
from pathlib import Path

from gotta.content.env import SESSION_ACTOR_ENV, SESSION_ENV, discover_state_env
from gotta.content.model import CommonOptions
from gotta.content.scope import (
    resolve_dirs,
    resolve_session_reference,
    session_identity,
    session_is_initialized,
    session_shared_id,
    session_surface_initialized,
)
from gotta import topology
from gotta.actor import (
    session_actor,
)

from .registry import (
    _actor_is_selected,
    _actor_registry,
    _actor_session_dir,
    _group_session_root,
    _resolve_bound_actor_name,
)


def _current_session_dir(
    explicit_session: str | None,
    explicit_actor: str | None = None,
    *,
    include_context_session: bool = True,
) -> Path | None:
    discovered = discover_state_env(include_context_session=include_context_session)
    try:
        dirs = resolve_dirs(
            CommonOptions(
                session_dir=explicit_session
                or os.environ.get(SESSION_ENV, "").strip()
                or discovered.get(SESSION_ENV, "").strip(),
                actor=explicit_actor
                or os.environ.get(SESSION_ACTOR_ENV, "").strip()
                or discovered.get(SESSION_ACTOR_ENV, "").strip(),
            ),
            create=False,
        )
    except Exception:
        return None
    return dirs.session_dir


def _session_dir(
    *,
    explicit_session: str | None,
    explicit_actor: str | None = None,
) -> Path:
    current = (
        resolve_session_reference(explicit_session, allow_missing=False)
        if explicit_session
        else _current_session_dir(
            explicit_session,
            explicit_actor=None,
        )
    )
    if current is None:
        raise SystemExit(
            "start or bind a session first with `gotta ...`. Stable interactive "
            "contexts adopt and scaffold their deterministic session on first "
            'session-aware use. Use `gotta session init --session "$WS"` only '
            "when you intentionally want to scaffold one exact root."
        )
    if explicit_actor:
        session_root = _group_session_root(current)
        current = _actor_session_dir(
            session_root,
            _resolve_bound_actor_name(session_root, explicit_actor),
        )
    elif (
        explicit_session
        and (current / "actors").is_dir()
        and topology.parse_shared_session_root(current) is not None
    ):
        current = _actor_session_dir(current, _require_primary_actor_name(current))
    if not session_is_initialized(current):
        raise SystemExit(
            "start or bind a session first with `gotta ...`. Stable interactive "
            "contexts adopt and scaffold their deterministic session on first "
            'session-aware use. Use `gotta session init --session "$WS"` only '
            "when you intentionally want to scaffold one exact root."
        )
    if not session_surface_initialized(current):
        raise SystemExit(
            "this session exists but is not scaffolded yet. In stable interactive "
            "contexts, first session-aware use should scaffold automatically. "
            "Use `gotta session init` only when you intentionally want to "
            "scaffold one exact root."
        )
    return current


def _primary_actor_name(work_dir: Path) -> str | None:
    session_root = _group_session_root(work_dir)
    selected = list(_selected_actor_ids(session_root))
    if not selected:
        return None
    shared_id = topology.normalize_identity(session_shared_id(session_root))
    if shared_id and shared_id in selected:
        return shared_id
    if len(selected) == 1:
        return selected[0]
    return None


def _require_primary_actor_name(work_dir: Path) -> str:
    primary = _primary_actor_name(work_dir)
    if primary:
        return primary
    raise SystemExit(
        "this shared session does not resolve to one canonical actor root; "
        "pass `--actor <actor>` explicitly"
    )


def _shared_session_read_root(raw: str) -> Path | None:
    resolved = resolve_session_reference(raw, allow_missing=False)
    if resolved is None:
        return None
    return _group_session_root(resolved)


def _shared_session_dir(
    *,
    explicit_session: str | None,
) -> Path:
    current = (
        _shared_session_read_root(explicit_session)
        if explicit_session
        else _current_session_dir(
            explicit_session,
            explicit_actor=None,
        )
    )
    if current is None:
        raise SystemExit(
            "start or bind a session first with `gotta ...`. Stable interactive "
            "contexts adopt and scaffold their deterministic session on first "
            'session-aware use. Use `gotta session init --session "$WS"` only '
            "when you intentionally want to scaffold one exact root."
        )
    return _group_session_root(current)


def _read_scope(
    *,
    explicit_session: str | None,
) -> tuple[Path, str]:
    current = _session_dir(
        explicit_session=explicit_session,
        explicit_actor=None,
    ).resolve()
    if (
        current.parent.name == "actors"
        or topology.parse_grouped_session_root(current) is not None
    ):
        actor_name = session_identity(current)
        if actor_name:
            return current, actor_name
    return current, ""


def _observation_scope(
    *,
    explicit_session: str | None,
    explicit_actor: str | None = None,
) -> tuple[Path, str]:
    if explicit_actor:
        current = _session_dir(
            explicit_session=explicit_session,
            explicit_actor=explicit_actor,
        ).resolve()
        return current, session_identity(current)
    current = _session_dir(
        explicit_session=explicit_session,
        explicit_actor=None,
    ).resolve()
    grouped_root = _group_session_root(current)
    if grouped_root != current:
        return grouped_root, ""
    return current, session_identity(current) if current.parent.name == "actors" else ""


def _selected_actor_ids(work_dir: Path) -> tuple[str, ...]:
    registry = _actor_registry(work_dir)
    selected: list[str] = []
    for actor_id in registry:
        if _actor_is_selected(work_dir, actor_id):
            selected.append(actor_id)
    return tuple(selected)


def _target_actor_ids(work_dir: Path, actor_ref: str | None = None) -> tuple[str, ...]:
    selected = _selected_actor_ids(work_dir)
    if not selected and session_is_initialized(work_dir):
        rooted_actor = session_actor(work_dir)
        if rooted_actor:
            selected = (rooted_actor,)
    if actor_ref:
        resolved = _resolve_bound_actor_name(work_dir, actor_ref)
        if resolved not in selected:
            raise SystemExit(
                f"{resolved} is not bound for this session; bind them first with "
                f"`gotta actor bind {actor_ref}`"
            )
        return (resolved,)
    return selected



"""Session, artifact, and local-path lookup for read resolution."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Any
import urllib.parse

from gotta import topology
from gotta.content.env import CONTENT_ENV, SESSION_ENV, load_state_env_at_root
from gotta.content.model import CommonOptions
from gotta.content.path import artifact_locator, is_sha256_digest, sanitize_name
from gotta.content.store import scan_content_store


def nearby_session_context() -> tuple[str, str]:
    session_root = os.environ.get(SESSION_ENV, "").strip()
    content_root = os.environ.get(CONTENT_ENV, "").strip()
    if session_root and content_root:
        return session_root, content_root
    cwd = Path.cwd().resolve()
    for parent in (cwd, *cwd.parents):
        data = load_state_env_at_root(parent)
        if not data:
            continue
        session_root = session_root or str(data.get(SESSION_ENV, "")).strip()
        content_root = content_root or str(data.get(CONTENT_ENV, "")).strip()
        if session_root or content_root:
            break
    if session_root and not content_root:
        content_root = str((Path(session_root).expanduser() / "content").resolve())
    return session_root, content_root


def explicit_session_context(options: CommonOptions | Any | None) -> tuple[str, str]:
    if options is None:
        return "", ""
    session_root = str(getattr(options, "session_dir", "") or "").strip()
    content_root = str(getattr(options, "content_dir", "") or "").strip()
    if session_root and not content_root:
        session_path = Path(session_root).expanduser().resolve()
        in_shared_topology = (
            topology.parse_grouped_session_root(session_path) is not None
            or topology.parse_shared_session_root(session_path) is not None
        )
        if in_shared_topology:
            shared_id = topology.shared_session_id(session_path)
            content_root = str(
                (topology.shared_session_root_for(shared_id) / "content").resolve()
            )
        else:
            content_root = str((session_path / "content").resolve())
    return session_root, content_root


def resolve_local_target(
    target: str,
    *,
    session_root: str = "",
    content_root: str = "",
) -> Path | None:
    candidate = Path(target).expanduser()
    session_root = session_root or nearby_session_context()[0]
    content_root = content_root or nearby_session_context()[1]
    digest_target = (
        target.removeprefix("content:") if target.startswith("content:") else target
    )
    if content_root and is_sha256_digest(digest_target):
        digest_candidate = (
            Path(content_root).expanduser() / digest_target / "data"
        ).resolve()
        if digest_candidate.exists():
            return digest_candidate
    if candidate.is_absolute():
        return candidate.resolve() if candidate.exists() else None
    if session_root:
        session_candidate = (Path(session_root).expanduser() / target).resolve()
        if session_candidate.exists():
            return session_candidate
    return None


def expected_local_target(target: str, *, session_root: str = "") -> Path | None:
    candidate = Path(target).expanduser()
    session_root = session_root or nearby_session_context()[0]
    if candidate.is_absolute():
        return candidate.resolve()
    if session_root:
        return (Path(session_root).expanduser() / target).resolve()
    return None


def resolve_session_artifact_name(
    target: str,
    *,
    content_root: str = "",
) -> Path | None:
    content_root = content_root or nearby_session_context()[1]
    if not content_root:
        return None
    root = Path(content_root).expanduser()
    if not root.exists():
        return None
    matches = [
        snapshot for snapshot in scan_content_store(root) if target in snapshot.names
    ]
    if not matches:
        return None
    if len(matches) > 1:
        suggestions = ", ".join(
            artifact_locator(target, snapshot.digest) for snapshot in matches[:5]
        )
        raise SystemExit(
            f"ambiguous stored artifact name '{target}' in the active session content store; "
            f"use `artifact:{sanitize_name(target)}@<digest12>` or `content:<digest>` instead. "
            f"matching artifact locators: {suggestions}"
        )
    return matches[0].data_path


def resolve_artifact_locator(target: str, *, content_root: str = "") -> Path | None:
    if not target.startswith("artifact:"):
        return None
    content_root = content_root or nearby_session_context()[1]
    if not content_root:
        raise SystemExit(
            f"artifact locator '{target}' requires an active or discoverable session content store"
        )
    root = Path(content_root).expanduser()
    if not root.exists():
        return None
    raw = target.removeprefix("artifact:")
    if "@" in raw:
        name_part, digest_hint = raw.rsplit("@", 1)
    else:
        name_part, digest_hint = raw, ""
    desired_name = sanitize_name(name_part)
    matches = [
        snapshot
        for snapshot in scan_content_store(root)
        if desired_name in snapshot.names
        and (not digest_hint or snapshot.digest.startswith(digest_hint))
    ]
    if not matches:
        return None
    if len(matches) > 1:
        suggestions = ", ".join(
            artifact_locator(desired_name, snapshot.digest) for snapshot in matches[:5]
        )
        raise SystemExit(
            f"ambiguous artifact locator '{target}' in the active session content store; "
            f"disambiguate with one of: {suggestions}"
        )
    return matches[0].data_path


def url_name(target: str) -> str:
    parsed = urllib.parse.urlparse(target)
    name = Path(parsed.path.rstrip("/")).name or parsed.netloc or "read"
    if "." not in name:
        name = f"{name}.md"
    return name
